from typing import Union

import pytdbot


class FileBoundMethods:
    def __init__(self):
        self._client: pytdbot.Client

    async def download(
        self,
        priority: int = 1,
        offset: int = 0,
        limit: int = 0,
        synchronous: bool = True,
    ) -> Union["pytdbot.types.Error", "pytdbot.types.File"]:
        r"""Downloads a file. Shortcut for :meth:`~pytdbot.Client.downloadFile`"""

        file_id = None
        if isinstance(self, pytdbot.types.RemoteFile):
            file_info = await self._client.getRemoteFile(self.id)
            if not file_info:
                return file_info

            file_id = file_info.id
        elif isinstance(self, pytdbot.types.File):
            file_id = self.id

        if file_id:
            return await self._client.downloadFile(
                file_id=file_id,
                priority=priority,
                offset=offset,
                limit=limit,
                synchronous=synchronous,
            )



from typing import List, Union
from functools import lru_cache

import pytdbot


class CallbackQueryBoundMethods:
    def __init__(self):
        self._client: pytdbot.Client

    @property
    @lru_cache(1)
    def callback_data(self) -> "pytdbot.utils.CallbackData":
        r"""Callback data that was set by :func:`~pytdbot.utils.callback_data`

        Args:
            action (``Any``):
                Action name

            data (``Any``, *optional*):
                Callback data
        """

        if isinstance(self.payload, pytdbot.types.CallbackQueryPayloadData):
            return pytdbot.utils.load_callback_data(self.payload.data)

        return pytdbot.utils.empty_callback_data

    @property
    @lru_cache(1)
    def text(self) -> str:
        r"""Callback data decoded as str"""

        if isinstance(self.payload, pytdbot.types.CallbackQueryPayloadData):
            return self.payload.data.decode("utf-8")

        return ""

    async def getMessage(self) -> Union["pytdbot.types.Error", "pytdbot.types.Message"]:
        r"""Get callback query message"""

        if self.message_id:
            return await self._client.getMessage(
                chat_id=self.chat_id, message_id=self.message_id
            )

    async def answer(
        self,
        text: str,
        show_alert: bool = None,
        url: str = None,
        cache_time: int = None,
    ) -> Union["pytdbot.types.Error", "pytdbot.types.Ok"]:
        r"""Answer to callback query. Shortcut for :meth:`~pytdbot.Client.answerCallbackQuery`"""

        return await self._client.answerCallbackQuery(
            self.id, text=text, show_alert=show_alert, url=url, cache_time=cache_time
        )

    async def edit_message_text(
        self,
        text: str,
        parse_mode: str = None,
        entities: list = None,
        disable_web_page_preview: bool = False,
        url: str = None,
        force_small_media: bool = None,
        force_large_media: bool = None,
        show_above_text: bool = None,
        reply_markup: "pytdbot.types.ReplyMarkup" = None,
    ) -> Union["pytdbot.types.Error", "pytdbot.types.Message"]:
        r"""Edit callback query message text. Shortcut for :meth:`~pytdbot.Client.editTextMessage`"""

        return await self._client.editTextMessage(
            chat_id=self.chat_id,
            message_id=self.message_id,
            text=text,
            parse_mode=parse_mode,
            entities=entities,
            disable_web_page_preview=disable_web_page_preview,
            url=url,
            force_small_media=force_small_media,
            force_large_media=force_large_media,
            show_above_text=show_above_text,
            reply_markup=reply_markup,
        )

    async def edit_message_caption(
        self,
        caption: str,
        caption_entities: List["pytdbot.types.TextEntity"] = None,
        parse_mode: str = None,
        show_caption_above_media: bool = None,
        reply_markup: "pytdbot.types.ReplyMarkup" = None,
    ) -> Union["pytdbot.types.Error", "pytdbot.types.Message"]:
        r"""Edit message caption"""

        parse_mode = parse_mode or self._client.default_parse_mode
        if isinstance(caption_entities, list):
            caption = pytdbot.types.FormattedText(
                text=caption, entities=caption_entities
            )
        elif parse_mode and isinstance(parse_mode, str):
            parse = await self._client.parseText(caption, parse_mode=parse_mode)
            if isinstance(parse, pytdbot.types.Error):
                return parse
            caption = parse
        else:
            caption = pytdbot.types.FormattedText(caption)

        return await self._client.editMessageCaption(
            chat_id=self.chat_id,
            message_id=self.message_id,
            caption=caption,
            show_caption_above_media=show_caption_above_media,
            reply_markup=reply_markup,
        )

    # TODO: edit_message_media?

    async def edit_message_reply_markup(
        self, reply_markup: "pytdbot.types.ReplyMarkup"
    ) -> Union["pytdbot.types.Error", "pytdbot.types.Message"]:
        r"""Edit message reply markup. Shortcut for :meth:`~pytdbot.Client.editMessageReplyMarkup`"""

        return await self._client.editMessageReplyMarkup(
            chat_id=self.chat_id, message_id=self.message_id, reply_markup=reply_markup
        )



import asyncio


def get_running_loop():
    """Get the current running event loop, or create a new one if none is running"""

    try:
        return asyncio.get_running_loop()
    except RuntimeError:
        return asyncio.new_event_loop()



from asyncio import sleep
from typing import Literal

import pytdbot


class ChatActions:
    def __init__(
        self,
        client: "pytdbot.Client",
        chat_id: int,
        action: Literal[
            "typing",
            "upload_photo",
            "record_video",
            "upload_video",
            "record_voice",
            "upload_voice",
            "upload_document",
            "choose_sticker",
            "find_location",
            "record_video_note",
            "upload_video_note",
            "cancel",
        ],
        message_thread_id: int = None,
    ) -> None:
        self.client = client
        self.chat_id = chat_id
        self.action = None
        self.__task = None
        self.message_thread_id = message_thread_id or 0

        assert isinstance(self.message_thread_id, int), "message_thread_id must be int"
        self.setAction(action)

    def __await__(self):
        return self.sendAction().__await__()

    async def __aenter__(self):
        await self.sendAction()
        self.__task = self.client.loop.create_task(self.__loop_action())
        return self

    async def __aexit__(self, exc_type, exc, traceback):
        await self.stop()

    async def sendAction(self):
        return await self.client.sendChatAction(
            chat_id=self.chat_id,
            message_thread_id=self.message_thread_id,
            action=self.action,
        )

    def setAction(
        self,
        action: Literal[
            "typing",
            "upload_photo",
            "record_video",
            "upload_video",
            "record_voice",
            "upload_voice",
            "upload_document",
            "choose_sticker",
            "find_location",
            "record_video_note",
            "upload_video_note",
            "cancel",
        ],
    ):
        if action == "typing" or action == "chatActionTyping":
            self.action = pytdbot.types.ChatActionTyping()
        elif action == "upload_photo" or action == "chatActionUploadingPhoto":
            self.action = pytdbot.types.ChatActionUploadingPhoto()
        elif action == "record_video" or action == "chatActionRecordingVideo":
            self.action = pytdbot.types.ChatActionRecordingVideo()
        elif action == "upload_video" or action == "chatActionUploadingVideo":
            self.action = pytdbot.types.ChatActionUploadingVideo()
        elif action == "record_voice" or action == "chatActionRecordingVoiceNote":
            self.action = pytdbot.types.ChatActionRecordingVoiceNote()
        elif action == "upload_voice" or action == "chatActionUploadingVoiceNote":
            self.action = pytdbot.types.ChatActionUploadingVoiceNote()
        elif action == "upload_document" or action == "chatActionUploadingDocument":
            self.action = pytdbot.types.ChatActionUploadingDocument()
        elif action == "choose_sticker" or action == "chatActionChoosingSticker":
            self.action = pytdbot.types.ChatActionChoosingSticker()
        elif action == "find_location" or action == "chatActionChoosingLocation":
            self.action = pytdbot.types.ChatActionChoosingLocation()
        elif action == "record_video_note" or action == "chatActionRecordingVideoNote":
            self.action = pytdbot.types.ChatActionRecordingVideoNote()
        elif action == "upload_video_note" or action == "chatActionUploadingVideoNote":
            self.action = pytdbot.types.ChatActionUploadingVideoNote()
        elif action == "cancel" or action == "chatActionCancel":
            self.action = pytdbot.types.ChatActionCancel()
        else:
            raise ValueError(f"Unknown action type {action}")

    async def __loop_action(self):
        while True:
            await sleep(4)
            await self.sendAction()

    async def stop(self):
        self.setAction("cancel")
        self.__task.cancel()
        await self.sendAction()



import asyncio
import logging
from typing import List
from concurrent.futures import ThreadPoolExecutor

import pytdbot
from .tdjson import TdJson

logger = logging.getLogger(__name__)


class ClientManager:
    """Client manager for Pytdbot"""

    def __init__(
        self,
        clients: List["pytdbot.Client"] = None,
        lib_path: str = None,
        verbosity: int = 2,
        loop: asyncio.AbstractEventLoop = None,
    ) -> None:
        """Manage multiple Pytdbot clients

        Example:
            .. code-block:: python

                >>> from pytdbot import ClientManager, Client
                >>> clients = [Client(...), Client(...), Client(...)]
                >>> client_manager = ClientManager(clients)
                >>> await client_manager.start()

        Parameters:
            clients (``List[pytdbot.Client]``, *optional*):
                List of clients to manage

            lib_path (``str``, *optional*):
                Path to TDlib library

            verbosity (``int``, *optional*):
                Verbosity level of TDlib. Default is ``2``

            loop (``asyncio.AbstractEventLoop``, *optional*):
                Event loop to use
        """

        if clients and not isinstance(clients, (list, pytdbot.Client)):
            raise TypeError("clients must be a list of pytdbot.Client")

        self.loop = loop or pytdbot.utils.get_running_loop()
        self.__tdjson = TdJson(lib_path, verbosity)

        self.__clients: dict[int, pytdbot.Client] = {}

        if isinstance(clients, list):
            self.__pending_clients = clients
            self.start_clients_on_add = True
        elif isinstance(clients, pytdbot.Client):
            self.__pending_clients = [clients]
            self.start_clients_on_add = False
        else:
            self.__pending_clients = None
            self.start_clients_on_add = False

        self.__receiver_task = None
        self.__should_exit = False
        self.is_running = False

    async def start(self) -> None:
        """Start the Client Manager"""

        if self.is_running:
            return

        self.__receiver_task = self.loop.create_task(self.__td_receiver_loop())

        for client in self.__pending_clients:
            await self.add_client(client, start_client=self.start_clients_on_add)

        self.__pending_clients = None

    async def add_client(
        self, client: "pytdbot.Client", start_client: bool = False
    ) -> None:
        """Add a client to the manager

        Parameters:
            client (``pytdbot.Client``):
                Client to add

            start_client (``bool``, *optional*):
                Whether to start the client immediately. Default is ``False``
        """

        if not isinstance(client, pytdbot.Client):
            raise TypeError("client must be an instance of pytdbot.Client")

        client_id = self.__tdjson.create_client_id()
        client.client_id = client_id
        client.client_manager = self

        self.__clients[client_id] = client

        if start_client:
            await client.start()

        logger.debug(f"Client {client_id} added")

    async def delete_client(self, client_id: int, close_client: bool = False) -> None:
        """Remove a client from the manager

        Parameters:
            client_id (``int``):
                ID of client to remove

            close_client (``bool``, *optional*):
                Whether to close the client before removing. Default is ``False``
        """

        client = self.__clients.pop(client_id, None)
        if not client:
            raise ValueError(f"Client with ID {client_id} not found")

        if close_client:
            await client.stop()
        logger.debug(f"Client {client_id} deleted")

    def send(self, client_id: int, request: dict) -> None:
        """Send a request to TDlib

        Parameters:
            client_id (``int``):
                ID of client to send request from

            request (``dict``):
                Request to send
        """

        self.__tdjson.send(client_id, request)

    async def __td_receiver_loop(self) -> None:
        with ThreadPoolExecutor(
            max_workers=1, thread_name_prefix="ClientManager"
        ) as executor:
            try:
                self.is_running = True
                logger.info("ClientManager started")

                while not self.__should_exit:
                    update = await self.loop.run_in_executor(
                        executor,
                        self.__tdjson.receive,
                        100000.0,  # Seconds
                    )

                    if not update or self.__should_exit:
                        continue

                    client = self.__clients.get(update["@client_id"])
                    if client:
                        self.loop.create_task(client.process_update(update))
                    else:
                        logger.warning(
                            f"Unknown client ID in update: {update['@client_id']}"
                        )

            except Exception:
                logger.exception("Error in td_receiver")
            finally:
                self.is_running = False
                logger.debug("ClientManager stopped")

    async def close(self, close_all_clients: bool = False) -> bool:
        """Close the Client Manager

        Parameters:
            close_all_clients (``bool``, *optional*):
                Whether to close all managed clients. Default is ``False``

        Returns:
            ``bool``
        """
        if self.__should_exit:
            return True

        self.__should_exit = True

        if close_all_clients:
            for client_id in list(self.__clients.keys()):
                await self.delete_client(client_id, close_client=True)

        # Send dummy request to wake up receiver
        self.send(0, {"@type": "getOption", "name": "version"})

        await self.__receiver_task

        logger.info("ClientManager closed")
        return True



import pytdbot
from typing import Literal, Union


class ScheduledEvent:
    r"""Describes a scheduled event

    Parameters:
        event_id (:class:`int`):
            Unique identifier of the scheduled event

        send_at (:class:`int`):
            Point in time \(Unix timestamp\) when the scheduled event will be sent
    """

    def __init__(
        self,
        event_id: int = 0,
        send_at: int = 0,
    ) -> None:
        self.event_id = event_id
        r"""Unique identifier of the scheduled event"""
        self.send_at = send_at
        r"""Point in time \(Unix timestamp\) when the scheduled event will be sent"""

    def __str__(self):
        return str(pytdbot.utils.obj_to_json(self, indent=4))

    def getType(self) -> Literal["scheduledEvent"]:
        return "scheduledEvent"

    def getClass(self) -> Literal["ScheduledEvent"]:
        return "ScheduledEvent"

    def to_dict(self) -> dict:
        return {
            "@type": self.getType(),
            "event_id": self.event_id,
            "send_at": self.send_at,
        }

    @classmethod
    def from_dict(cls, data: dict) -> Union["ScheduledEvent", None]:
        if data:
            data_class = cls()
            data_class.event_id = data.get("event_id", None)
            data_class.send_at = data.get("send_at", None)

        return data_class


class UpdateScheduledEvent:
    r"""A scheduled event

    Parameters:
        name (:class:`str`):
            Event name

        event_id (:class:`int`):
            Unique identifier of the scheduled event

        payload (:class:`str`):
            Event payload
    """

    def __init__(
        self,
        name: str = "",
        event_id: int = 0,
        payload: str = "",
    ) -> None:
        self.name = name
        r"""Event name"""
        self.event_id = event_id
        r"""Unique identifier of the scheduled event"""
        self.payload = payload
        r"""Event payload"""

    def __str__(self):
        return str(pytdbot.utils.obj_to_json(self, indent=4))

    def getType(self) -> Literal["updateScheduledEvent"]:
        return "updateScheduledEvent"

    def getClass(self) -> Literal["UpdateScheduledEvent"]:
        return "UpdateScheduledEvent"

    def to_dict(self) -> dict:
        return {
            "@type": self.getType(),
            "name": self.name,
            "event_id": self.event_id,
            "payload": self.payload,
        }

    @classmethod
    def from_dict(cls, data: dict) -> Union["UpdateScheduledEvent", None]:
        if data:
            data_class = cls()
            data_class.name = data.get("name", None)
            data_class.event_id = data.get("event_id", None)
            data_class.payload = data.get("payload", None)

        return data_class



from functools import lru_cache
from typing import List, Literal, Union

import pytdbot

from .chatActions import ChatActions


class MessageBoundMethods:
    def __init__(self):
        self._client: pytdbot.Client

    @property
    @lru_cache(1)
    def from_id(self) -> Union[int, None]:
        r"""Message Sender ID"""

        if isinstance(self.sender_id, pytdbot.types.MessageSenderChat):
            return self.sender_id.chat_id
        elif isinstance(self.sender_id, pytdbot.types.MessageSenderUser):
            return self.sender_id.user_id

    @property
    @lru_cache(1)
    def reply_to_message_id(self) -> int:
        r"""Replied message ID"""

        if isinstance(self.reply_to, pytdbot.types.MessageReplyToMessage):
            return self.reply_to.message_id

    @property
    @lru_cache(1)
    def text(self) -> str:
        r"""Text of the message"""

        if isinstance(self.content, pytdbot.types.MessageText):
            return self.content.text.text

        return ""

    @property
    @lru_cache(1)
    def entities(self) -> Union[List["pytdbot.types.TextEntity"], None]:
        r"""Entities of the message"""

        if isinstance(self.content, pytdbot.types.MessageText):
            return self.content.text.entities

    @property
    @lru_cache(1)
    def caption(self) -> Union[str, None]:
        r"""Caption of the received media"""

        if isinstance(
            self.content,
            (
                pytdbot.types.MessagePhoto,
                pytdbot.types.MessageVideo,
                pytdbot.types.MessageAnimation,
                pytdbot.types.MessageAudio,
                pytdbot.types.MessageDocument,
                pytdbot.types.MessageVoiceNote,
            ),
        ):
            return self.content.caption.text

    @property
    @lru_cache(1)
    def caption_entities(self) -> Union[List["pytdbot.types.TextEntity"], None]:
        r"""Caption entities of the received media"""

        if isinstance(
            self.content,
            (
                pytdbot.types.MessagePhoto,
                pytdbot.types.MessageVideo,
                pytdbot.types.MessageAnimation,
                pytdbot.types.MessageAudio,
                pytdbot.types.MessageDocument,
                pytdbot.types.MessageVoiceNote,
            ),
        ):
            return self.content.caption.entities

    @property
    @lru_cache(1)
    def remote_file_id(self) -> Union[str, None]:
        r"""Remote file id"""

        file_id = None
        if isinstance(self.content, pytdbot.types.MessagePhoto):
            file_id = self.content.photo.sizes[-1].photo.remote.id
        elif isinstance(self.content, pytdbot.types.MessageVideo):
            file_id = self.content.video.video.remote.id
        elif isinstance(self.content, pytdbot.types.MessageSticker):
            file_id = self.content.sticker.sticker.remote.id
        elif isinstance(self.content, pytdbot.types.MessageAnimation):
            file_id = self.content.animation.animation.remote.id
        elif isinstance(self.content, pytdbot.types.MessageAudio):
            file_id = self.content.audio.audio.remote.id
        elif isinstance(self.content, pytdbot.types.MessageDocument):
            file_id = self.content.document.document.remote.id
        elif isinstance(self.content, pytdbot.types.MessageVoiceNote):
            file_id = self.content.voice_note.voice.remote.id
        elif isinstance(self.content, pytdbot.types.MessageVideoNote):
            file_id = self.content.video_note.video.remote.id

        return file_id

    @property
    @lru_cache(1)
    def remote_unique_file_id(self) -> Union[str, None]:
        r"""Remote unique file id"""

        unique_file_id = None
        if isinstance(self.content, pytdbot.types.MessagePhoto):
            unique_file_id = self.content.photo.sizes[-1].photo.remote.unique_id
        elif isinstance(self.content, pytdbot.types.MessageVideo):
            unique_file_id = self.content.video.video.remote.unique_id
        elif isinstance(self.content, pytdbot.types.MessageSticker):
            unique_file_id = self.content.sticker.sticker.remote.unique_id
        elif isinstance(self.content, pytdbot.types.MessageAnimation):
            unique_file_id = self.content.animation.animation.remote.unique_id
        elif isinstance(self.content, pytdbot.types.MessageAudio):
            unique_file_id = self.content.audio.audio.remote.unique_id
        elif isinstance(self.content, pytdbot.types.MessageDocument):
            unique_file_id = self.content.document.document.remote.unique_id
        elif isinstance(self.content, pytdbot.types.MessageVoiceNote):
            unique_file_id = self.content.voice_note.voice.remote.unique_id
        elif isinstance(self.content, pytdbot.types.MessageVideoNote):
            unique_file_id = self.content.video_note.video.remote.unique_id

        return unique_file_id

    async def mention(self, parse_mode: str = "html") -> Union[str, None]:
        r"""Get the text_mention of the message sender

        Parameters:
            parse_mode (``str``, *optional*):
                The parse mode of the mention. Default is ``html``
        """

        chat = await self._client.getChat(self.from_id)
        if chat:
            return pytdbot.utils.mention(
                chat.title,
                self.from_id,
                html=True if parse_mode.lower() == "html" else False,
            )

    async def getMessageProperties(
        self,
    ) -> Union["pytdbot.types.Error", "pytdbot.types.MessageProperties"]:
        r"""Get the message properties"""

        return await self._client.getMessageProperties(
            chat_id=self.chat_id,
            message_id=self.id,
        )

    async def getMessageLink(
        self,
        media_timestamp: int = 0,
        for_album: bool = False,
        in_message_thread: bool = False,
    ) -> Union["pytdbot.types.Error", "pytdbot.types.MessageLink"]:
        r"""Get message link

        Parameters:
            media_timestamp (:class:`int`):
                If not 0, timestamp from which the video/audio/video note/voice note/story playing must start, in seconds\. The media can be in the message content or in its link preview

            for_album (:class:`bool`):
                Pass true to create a link for the whole media album

            in_message_thread (:class:`bool`):
                Pass true to create a link to the message as a channel post comment, in a message thread, or a forum topic
        """

        return await self._client.getMessageLink(
            chat_id=self.chat_id,
            message_id=self.id,
            media_timestamp=media_timestamp,
            for_album=for_album,
            in_message_thread=in_message_thread,
        )

    async def getRepliedMessage(
        self,
    ) -> Union["pytdbot.types.Error", "pytdbot.types.Message"]:
        r"""Get the replied message"""

        return await self._client.getRepliedMessage(
            chat_id=self.chat_id,
            message_id=self.id,
        )

    async def getChat(
        self,
    ) -> Union["pytdbot.types.Error", "pytdbot.types.Chat"]:
        r"""Get chat info"""

        return await self._client.getChat(self.chat_id)

    async def getChatMember(
        self,
    ) -> Union["pytdbot.types.Error", "pytdbot.types.ChatMember"]:
        r"""Get member info in the current chat"""

        return await self._client.getChatMember(
            chat_id=self.chat_id, member_id=self.sender_id
        )

    async def getUser(
        self,
    ) -> Union["pytdbot.types.Error", "pytdbot.types.User"]:
        r"""Get user info"""

        return await self._client.getUser(self.from_id)

    async def setChatMemberStatus(
        self,
        status: "pytdbot.types.ChatMemberStatus",
    ) -> Union["pytdbot.types.Error", "pytdbot.types.Ok"]:
        r"""Set chat member status"""

        return await self._client.setChatMemberStatus(
            chat_id=self.chat_id, member_id=self.sender_id, status=status
        )

    async def leaveChat(self) -> Union["pytdbot.types.Error", "pytdbot.types.Ok"]:
        r"""Leave the current chat"""

        return await self._client.leaveChat(self.chat_id)

    async def ban(
        self, banned_until_date: int = 0
    ) -> Union["pytdbot.types.Error", "pytdbot.types.Ok"]:
        r"""Ban the message sender

        Parameters:

            banned_until_date (``int``):
                Point in time (Unix timestamp) when the user will be unbanned; 0 if never. If the user is banned for more than 366 days or for less than 30 seconds from the current time, the user is considered to be banned forever. Always 0 in basic groups
        """

        return await self.setChatMemberStatus(
            status=pytdbot.types.ChatMemberStatusBanned(
                banned_until_date=banned_until_date
            )
        )

    async def delete(
        self,
        revoke: bool = True,
    ) -> Union["pytdbot.types.Error", "pytdbot.types.Ok"]:
        r"""Delete the received message

        Parameters:
            revoke (``bool``, *optional*):
                Pass true to delete messages for all chat members. Always true for supergroups, channels and secret chats
        """

        return await self._client.deleteMessages(
            chat_id=self.chat_id, message_ids=[self.id], revoke=revoke
        )

    async def react(
        self, emoji: str = "👍", is_big: bool = False
    ) -> Union["pytdbot.types.Error", "pytdbot.types.Ok"]:
        r"""React to the current message

        Parameters:
            emoji (``str``, *optional*):
                Text representation of the reaction; pass ``None`` to remove the current reaction. Default is ``👍``

            is_big (``bool``, *optional*):
                Pass true if the reactions are added with a big animation. Default is ``False``
        """

        return await self._client.setMessageReactions(
            chat_id=self.chat_id,
            message_id=self.id,
            reaction_types=None
            if not emoji
            else [pytdbot.types.ReactionTypeEmoji(emoji=emoji)],
            is_big=is_big,
        )

    async def pin(
        self,
        disable_notification: bool = False,
        only_for_self: bool = False,
    ) -> Union["pytdbot.types.Error", "pytdbot.types.Ok"]:
        r"""Pin the message

        Parameters:
            disable_notification (``bool``, *optional*):
                If True, disable notification for the message

            only_for_self (``bool``, *optional*):
                True, if the message needs to be pinned for one side only; private chats only
        """

        return await self._client.pinChatMessage(
            chat_id=self.chat_id,
            message_id=self.id,
            disable_notification=disable_notification,
            only_for_self=only_for_self,
        )

    async def unpin(self) -> Union["pytdbot.types.Error", "pytdbot.types.Ok"]:
        r"""Unpin the message"""

        return await self._client.unpinChatMessage(
            chat_id=self.chat_id, message_id=self.id
        )

    async def download(
        self,
        priority: int = 1,
        offset: int = 0,
        limit: int = 0,
        synchronous: bool = True,
    ) -> Union["pytdbot.types.Error", "pytdbot.types.LocalFile"]:
        r"""Download the media file and returns ``LocalFile`` object. Shortcut for :meth:`~pytdbot.Client.downloadFile`."""

        res = None
        if isinstance(self.content, pytdbot.types.MessagePhoto):
            res = await self.content.photo.sizes[-1].photo.download(
                priority=priority, offset=offset, limit=limit, synchronous=synchronous
            )
        elif isinstance(self.content, pytdbot.types.MessageVideo):
            res = await self.content.video.video.download(
                priority=priority, offset=offset, limit=limit, synchronous=synchronous
            )
        elif isinstance(self.content, pytdbot.types.MessageSticker):
            res = await self.content.sticker.sticker.download(
                priority=priority, offset=offset, limit=limit, synchronous=synchronous
            )
        elif isinstance(self.content, pytdbot.types.MessageAnimation):
            res = await self.content.animation.animation.download(
                priority=priority, offset=offset, limit=limit, synchronous=synchronous
            )
        elif isinstance(self.content, pytdbot.types.MessageAudio):
            res = await self.content.audio.audio.download(
                priority=priority, offset=offset, limit=limit, synchronous=synchronous
            )
        elif isinstance(self.content, pytdbot.types.MessageDocument):
            res = await self.content.document.document.download(
                priority=priority, offset=offset, limit=limit, synchronous=synchronous
            )
        elif isinstance(self.content, pytdbot.types.MessageVoiceNote):
            res = await self.content.voice_note.voice.download(
                priority=priority, offset=offset, limit=limit, synchronous=synchronous
            )
        elif isinstance(self.content, pytdbot.types.MessageVideoNote):
            res = await self.content.video_note.video.download(
                priority=priority, offset=offset, limit=limit, synchronous=synchronous
            )

        if isinstance(res, pytdbot.types.Error):
            return res
        elif isinstance(res, pytdbot.types.File):
            return res.local

    def action(
        self,
        action: Literal[
            "typing",
            "upload_photo",
            "record_video",
            "upload_video",
            "record_voice",
            "upload_voice",
            "upload_document",
            "choose_sticker",
            "find_location",
            "record_video_note",
            "upload_video_note",
            "cancel",
        ],
        message_thread_id: int = None,
    ) -> ChatActions:
        r"""Sends a chat action to a specific chat. Supporting context manager (``with`` statement)

        \Example:


            .. code-block:: python

                async with update.action("record_video") as action:
                    ## Any blocking operation
                    await asyncio.sleep(10)
                    action.setAction("upload_video") # change the action to uploading a video

        Or


            .. code-block:: python

                await update.action("typing")
                ## Any blocking operation
                await asyncio.sleep(2)
                await update.reply_text("Hello?")

        \Parameters:
            action (``str``):
                Type of action to broadcast. Choose one, depending on what the user is about to receive: ``typing`` for text messages, ``upload_photo`` for photos, ``record_video`` or ``upload_video`` for videos, ``record_voice`` or ``upload_voice`` for voice notes, ``upload_document`` for general files, ``choose_sticker`` for stickers, ``find_location` for location data, ``record_video_note`` or ``upload_video_note`` for video notes

            message_thread_id (``int``, *optional*):
                If not 0, a message thread identifier in which the action was performed. Default is ``None``
        """

        return ChatActions(
            client=self._client,
            chat_id=self.chat_id,
            action=action,
            message_thread_id=message_thread_id,
        )

    async def reply_text(
        self,
        text: str,
        entities: List["pytdbot.types.TextEntity"] = None,
        parse_mode: str = None,
        disable_web_page_preview: bool = False,
        url: str = None,
        force_small_media: bool = None,
        force_large_media: bool = None,
        show_above_text: bool = None,
        clear_draft: bool = False,
        disable_notification: bool = False,
        protect_content: bool = False,
        allow_paid_broadcast: bool = False,
        message_thread_id: int = 0,
        quote: "pytdbot.types.InputTextQuote" = None,
        reply_markup: Union[
            "pytdbot.types.ReplyMarkupInlineKeyboard",
            "pytdbot.types.ReplyMarkupShowKeyboard",
            "pytdbot.types.ReplyMarkupForceReply",
            "pytdbot.types.ReplyMarkupRemoveKeyboard",
        ] = None,
    ) -> Union["pytdbot.types.Error", "pytdbot.types.Message"]:
        r"""Reply to the message with text. Shortcut for :meth:`~pytdbot.Client.sendTextMessage`."""

        return await self._client.sendTextMessage(
            chat_id=self.chat_id,
            text=text,
            entities=entities,
            parse_mode=parse_mode,
            disable_web_page_preview=disable_web_page_preview,
            url=url,
            force_small_media=force_small_media,
            force_large_media=force_large_media,
            show_above_text=show_above_text,
            clear_draft=clear_draft,
            disable_notification=disable_notification,
            protect_content=protect_content,
            allow_paid_broadcast=allow_paid_broadcast,
            message_thread_id=message_thread_id,
            quote=quote,
            reply_to_message_id=self.id,
            reply_markup=reply_markup,
        )

    async def reply_animation(
        self,
        animation: Union["pytdbot.types.InputFile", str],
        thumbnail: "pytdbot.types.InputThumbnail" = None,
        caption: str = None,
        caption_entities: list = None,
        parse_mode: str = None,
        added_sticker_file_ids: list = None,
        duration: int = 0,
        width: int = 0,
        height: int = 0,
        disable_notification: bool = False,
        protect_content: bool = False,
        allow_paid_broadcast: bool = False,
        has_spoiler: bool = False,
        message_thread_id: int = 0,
        quote: "pytdbot.types.InputTextQuote" = None,
        reply_markup: Union[
            "pytdbot.types.ReplyMarkupInlineKeyboard",
            "pytdbot.types.ReplyMarkupShowKeyboard",
            "pytdbot.types.ReplyMarkupForceReply",
            "pytdbot.types.ReplyMarkupRemoveKeyboard",
        ] = None,
    ) -> Union["pytdbot.types.Error", "pytdbot.types.Message"]:
        r"""Reply to the message with animation. Shortcut for :meth:`~pytdbot.Client.sendAnimation`."""

        return await self._client.sendAnimation(
            chat_id=self.chat_id,
            animation=animation,
            thumbnail=thumbnail,
            caption=caption,
            caption_entities=caption_entities,
            parse_mode=parse_mode,
            added_sticker_file_ids=added_sticker_file_ids,
            duration=duration,
            width=width,
            height=height,
            disable_notification=disable_notification,
            protect_content=protect_content,
            allow_paid_broadcast=allow_paid_broadcast,
            has_spoiler=has_spoiler,
            message_thread_id=message_thread_id,
            quote=quote,
            reply_to_message_id=self.id,
            reply_markup=reply_markup,
        )

    async def reply_audio(
        self,
        audio: Union["pytdbot.types.InputFile", str],
        album_cover_thumbnail: "pytdbot.types.InputThumbnail" = None,
        caption: str = None,
        caption_entities: list = None,
        parse_mode: str = None,
        title: str = None,
        performer: str = None,
        duration: int = 0,
        disable_notification: bool = False,
        protect_content: bool = False,
        allow_paid_broadcast: bool = False,
        message_thread_id: int = 0,
        quote: "pytdbot.types.InputTextQuote" = None,
        reply_markup: Union[
            "pytdbot.types.ReplyMarkupInlineKeyboard",
            "pytdbot.types.ReplyMarkupShowKeyboard",
            "pytdbot.types.ReplyMarkupForceReply",
            "pytdbot.types.ReplyMarkupRemoveKeyboard",
        ] = None,
    ) -> Union["pytdbot.types.Error", "pytdbot.types.Message"]:
        r"""Reply to the message with audio. Shortcut for :meth:`~pytdbot.Client.sendAudio`."""

        return await self._client.sendAudio(
            chat_id=self.chat_id,
            audio=audio,
            album_cover_thumbnail=album_cover_thumbnail,
            caption=caption,
            caption_entities=caption_entities,
            parse_mode=parse_mode,
            title=title,
            performer=performer,
            duration=duration,
            disable_notification=disable_notification,
            protect_content=protect_content,
            allow_paid_broadcast=allow_paid_broadcast,
            message_thread_id=message_thread_id,
            quote=quote,
            reply_to_message_id=self.id,
            reply_markup=reply_markup,
        )

    async def reply_document(
        self,
        document: Union["pytdbot.types.InputFile", str],
        thumbnail: "pytdbot.types.InputThumbnail" = None,
        caption: str = None,
        caption_entities: list = None,
        parse_mode: str = None,
        disable_content_type_detection: bool = True,
        disable_notification: bool = False,
        protect_content: bool = False,
        allow_paid_broadcast: bool = False,
        message_thread_id: int = 0,
        quote: "pytdbot.types.InputTextQuote" = None,
        reply_markup: Union[
            "pytdbot.types.ReplyMarkupInlineKeyboard",
            "pytdbot.types.ReplyMarkupShowKeyboard",
            "pytdbot.types.ReplyMarkupForceReply",
            "pytdbot.types.ReplyMarkupRemoveKeyboard",
        ] = None,
    ) -> Union["pytdbot.types.Error", "pytdbot.types.Message"]:
        r"""Reply to the message with a document. Shortcut for :meth:`~pytdbot.Client.sendDocument`."""

        return await self._client.sendDocument(
            chat_id=self.chat_id,
            document=document,
            thumbnail=thumbnail,
            caption=caption,
            caption_entities=caption_entities,
            parse_mode=parse_mode,
            disable_content_type_detection=disable_content_type_detection,
            disable_notification=disable_notification,
            protect_content=protect_content,
            allow_paid_broadcast=allow_paid_broadcast,
            message_thread_id=message_thread_id,
            quote=quote,
            reply_to_message_id=self.id,
            reply_markup=reply_markup,
        )

    async def reply_photo(
        self,
        photo: Union["pytdbot.types.InputFile", str],
        thumbnail: "pytdbot.types.InputThumbnail" = None,
        caption: str = None,
        caption_entities: list = None,
        parse_mode: str = None,
        added_sticker_file_ids: list = None,
        width: int = 0,
        height: int = 0,
        self_destruct_type: "pytdbot.types.MessageSelfDestructType" = None,
        disable_notification: bool = False,
        protect_content: bool = False,
        allow_paid_broadcast: bool = False,
        has_spoiler: bool = False,
        message_thread_id: int = 0,
        quote: "pytdbot.types.InputTextQuote" = None,
        reply_markup: Union[
            "pytdbot.types.ReplyMarkupInlineKeyboard",
            "pytdbot.types.ReplyMarkupShowKeyboard",
            "pytdbot.types.ReplyMarkupForceReply",
            "pytdbot.types.ReplyMarkupRemoveKeyboard",
        ] = None,
    ) -> Union["pytdbot.types.Error", "pytdbot.types.Message"]:
        r"""Reply to the message with a photo. Shortcut for :meth:`~pytdbot.Client.sendPhoto`."""

        return await self._client.sendPhoto(
            chat_id=self.chat_id,
            photo=photo,
            thumbnail=thumbnail,
            caption=caption,
            caption_entities=caption_entities,
            parse_mode=parse_mode,
            added_sticker_file_ids=added_sticker_file_ids,
            width=width,
            height=height,
            self_destruct_type=self_destruct_type,
            disable_notification=disable_notification,
            protect_content=protect_content,
            allow_paid_broadcast=allow_paid_broadcast,
            has_spoiler=has_spoiler,
            message_thread_id=message_thread_id,
            quote=quote,
            reply_to_message_id=self.id,
            reply_markup=reply_markup,
        )

    async def reply_video(
        self,
        video: Union["pytdbot.types.InputFile", str],
        thumbnail: "pytdbot.types.InputThumbnail" = None,
        caption: str = None,
        caption_entities: list = None,
        parse_mode: str = None,
        added_sticker_file_ids: list = None,
        supports_streaming: bool = None,
        duration: int = 0,
        width: int = 0,
        height: int = 0,
        self_destruct_type: "pytdbot.types.MessageSelfDestructType" = None,
        disable_notification: bool = False,
        protect_content: bool = False,
        allow_paid_broadcast: bool = False,
        has_spoiler: bool = False,
        message_thread_id: int = 0,
        quote: "pytdbot.types.InputTextQuote" = None,
        reply_markup: Union[
            "pytdbot.types.ReplyMarkupInlineKeyboard",
            "pytdbot.types.ReplyMarkupShowKeyboard",
            "pytdbot.types.ReplyMarkupForceReply",
            "pytdbot.types.ReplyMarkupRemoveKeyboard",
        ] = None,
    ) -> Union["pytdbot.types.Error", "pytdbot.types.Message"]:
        r"""Reply to the message with a video. Shortcut for :meth:`~pytdbot.Client.sendVideo`."""

        return await self._client.sendVideo(
            chat_id=self.chat_id,
            video=video,
            thumbnail=thumbnail,
            caption=caption,
            caption_entities=caption_entities,
            parse_mode=parse_mode,
            added_sticker_file_ids=added_sticker_file_ids,
            supports_streaming=supports_streaming,
            duration=duration,
            width=width,
            height=height,
            self_destruct_type=self_destruct_type,
            disable_notification=disable_notification,
            protect_content=protect_content,
            allow_paid_broadcast=allow_paid_broadcast,
            has_spoiler=has_spoiler,
            message_thread_id=message_thread_id,
            quote=quote,
            reply_to_message_id=self.id,
            reply_markup=reply_markup,
        )

    async def reply_video_note(
        self,
        video_note: Union["pytdbot.types.InputFile", str],
        thumbnail: "pytdbot.types.InputThumbnail" = None,
        duration: int = 0,
        length: int = 0,
        disable_notification: bool = False,
        protect_content: bool = False,
        allow_paid_broadcast: bool = False,
        message_thread_id: int = 0,
        quote: "pytdbot.types.InputTextQuote" = None,
        reply_markup: Union[
            "pytdbot.types.ReplyMarkupInlineKeyboard",
            "pytdbot.types.ReplyMarkupShowKeyboard",
            "pytdbot.types.ReplyMarkupForceReply",
            "pytdbot.types.ReplyMarkupRemoveKeyboard",
        ] = None,
    ) -> Union["pytdbot.types.Error", "pytdbot.types.Message"]:
        r"""Reply to the message with a video note. Shortcut for :meth:`~pytdbot.Client.sendVideoNote`."""

        return await self._client.sendVideoNote(
            chat_id=self.chat_id,
            video_note=video_note,
            thumbnail=thumbnail,
            duration=duration,
            length=length,
            disable_notification=disable_notification,
            protect_content=protect_content,
            allow_paid_broadcast=allow_paid_broadcast,
            message_thread_id=message_thread_id,
            quote=quote,
            reply_to_message_id=self.id,
            reply_markup=reply_markup,
        )

    async def reply_voice(
        self,
        voice: Union["pytdbot.types.InputFile", str],
        caption: str = None,
        caption_entities: list = None,
        parse_mode: str = None,
        duration: int = 0,
        waveform: bytes = None,
        disable_notification: bool = False,
        protect_content: bool = False,
        allow_paid_broadcast: bool = False,
        message_thread_id: int = 0,
        quote: "pytdbot.types.InputTextQuote" = None,
        reply_markup: Union[
            "pytdbot.types.ReplyMarkupInlineKeyboard",
            "pytdbot.types.ReplyMarkupShowKeyboard",
            "pytdbot.types.ReplyMarkupForceReply",
            "pytdbot.types.ReplyMarkupRemoveKeyboard",
        ] = None,
    ) -> Union["pytdbot.types.Error", "pytdbot.types.Message"]:
        r"""Reply to the message with a voice note. Shortcut for :meth:`~pytdbot.Client.sendVoice`."""

        return await self._client.sendVoice(
            chat_id=self.chat_id,
            voice=voice,
            caption=caption,
            caption_entities=caption_entities,
            parse_mode=parse_mode,
            duration=duration,
            waveform=waveform,
            disable_notification=disable_notification,
            protect_content=protect_content,
            allow_paid_broadcast=allow_paid_broadcast,
            message_thread_id=message_thread_id,
            quote=quote,
            reply_to_message_id=self.id,
            reply_markup=reply_markup,
        )

    async def reply_sticker(
        self,
        sticker: Union["pytdbot.types.InputFile", str],
        emoji: str = None,
        thumbnail: "pytdbot.types.InputThumbnail" = None,
        width: int = 0,
        height: int = 0,
        disable_notification: bool = False,
        protect_content: bool = False,
        allow_paid_broadcast: bool = False,
        message_thread_id: int = 0,
        quote: "pytdbot.types.InputTextQuote" = None,
        reply_markup: Union[
            "pytdbot.types.ReplyMarkupInlineKeyboard",
            "pytdbot.types.ReplyMarkupShowKeyboard",
            "pytdbot.types.ReplyMarkupForceReply",
            "pytdbot.types.ReplyMarkupRemoveKeyboard",
        ] = None,
    ) -> Union["pytdbot.types.Error", "pytdbot.types.Message"]:
        r"""Reply to the message with a sticker. Shortcut for :meth:`~pytdbot.Client.sendSticker`."""

        return await self._client.sendSticker(
            chat_id=self.chat_id,
            sticker=sticker,
            emoji=emoji,
            thumbnail=thumbnail,
            width=width,
            height=height,
            disable_notification=disable_notification,
            protect_content=protect_content,
            allow_paid_broadcast=allow_paid_broadcast,
            message_thread_id=message_thread_id,
            quote=quote,
            reply_to_message_id=self.id,
            reply_markup=reply_markup,
        )

    async def copy(
        self,
        chat_id: int,
        in_game_share: bool = None,
        replace_caption: bool = None,
        new_caption: str = None,
        new_caption_entities: list = None,
        parse_mode: str = None,
        disable_notification: bool = False,
        protect_content: bool = False,
        allow_paid_broadcast: bool = False,
        message_thread_id: int = 0,
        quote: "pytdbot.types.InputTextQuote" = None,
        reply_to_message_id: int = 0,
    ) -> Union["pytdbot.types.Error", "pytdbot.types.Message"]:
        r"""Copy message to chat. Shortcut for :meth:`~pytdbot.Client.sendCopy`."""

        return await self._client.sendCopy(
            chat_id=chat_id,
            from_chat_id=self.chat_id,
            message_id=self.id,
            in_game_share=in_game_share,
            replace_caption=replace_caption,
            new_caption=new_caption,
            new_caption_entities=new_caption_entities,
            parse_mode=parse_mode,
            disable_notification=disable_notification,
            protect_content=protect_content,
            allow_paid_broadcast=allow_paid_broadcast,
            message_thread_id=message_thread_id,
            quote=quote,
            reply_to_message_id=reply_to_message_id,
        )

    async def forward(
        self,
        chat_id: int,
        in_game_share: bool = False,
        disable_notification: bool = False,
    ) -> Union["pytdbot.types.Error", "pytdbot.types.Message"]:
        r"""Forward message to chat. Shortcut for :meth:`~pytdbot.Client.forwardMessage`."""

        return await self._client.forwardMessage(
            chat_id=chat_id,
            from_chat_id=self.chat_id,
            message_id=self.id,
            in_game_share=in_game_share,
            disable_notification=disable_notification,
        )

    async def edit_text(
        self,
        text: str,
        parse_mode: str = None,
        entities: list = None,
        disable_web_page_preview: bool = False,
        url: str = None,
        force_small_media: bool = None,
        force_large_media: bool = None,
        show_above_text: bool = None,
        reply_markup: "pytdbot.types.ReplyMarkup" = None,
    ) -> Union["pytdbot.types.Error", "pytdbot.types.Message"]:
        r"""Edit text message. Shortcut for :meth:`~pytdbot.Client.editTextMessage`."""

        return await self._client.editTextMessage(
            chat_id=self.chat_id,
            message_id=self.id,
            text=text,
            parse_mode=parse_mode,
            entities=entities,
            disable_web_page_preview=disable_web_page_preview,
            url=url,
            force_small_media=force_small_media,
            force_large_media=force_large_media,
            show_above_text=show_above_text,
            reply_markup=reply_markup,
        )



__all__ = ("Plugins",)


class Plugins:
    """Load and filter plugins from a folder"""

    def __init__(self, folder: str, include: list = None, exclude: list = None) -> None:
        """
        Parameters:
            folder (``str``):
                The folder to load plugins from

            include (``list``, *optional*):
                Only load plugins with names in this list

            exclude (``list``, *optional*):
                Exclude plugins with names in this list

        Example:
            To load only the plugins with path "plugins/rules.py" and "plugins/subfolder1/commands.py",
            you should create the ``Plugins`` object like this:

            >>> plugins = Plugins(
                    folder="plugins/",
                    include=[
                        "rules" # will be translated to "plugins.rules"
                        "subfolder1.commands" # -> plugins.subfolder1.commands
                    ]
                )
        Raises:
            TypeError
        """

        if not isinstance(folder, str):
            raise TypeError("folder must be str")
        elif include is not None and not isinstance(include, list):
            raise TypeError("include must be list or None")
        elif exclude is not None and not isinstance(exclude, list):
            raise TypeError("include must be list or None")

        self.folder = folder
        self.include = include
        self.exclude = exclude



from re import findall

from setuptools import find_packages, setup

with open("pytdbot/__init__.py", "r") as f:
    version = findall(r"__version__ = \"(.+)\"", f.read())[0]

with open("README.md", "r") as f:
    readme = f.read()

with open("requirements.txt", "r") as f:
    requirements = [x.strip() for x in f.readlines()]


setup(
    name="Pytdbot",
    version=version,
    description="Easy-to-use asynchronous TDLib wrapper for Python.",
    long_description=readme,
    long_description_content_type="text/markdown",
    author="AYMEN Mohammed",
    author_email="let.me.code.safe@gmail.com",
    url="https://github.com/pytdbot/client",
    license="MIT",
    python_requires=">=3.9",
    install_requires=requirements,
    extras_require={
        "tdjson": ["tdjson"],
    },
    project_urls={
        "Source": "https://github.com/pytdbot/client",
        "Tracker": "https://github.com/pytdbot/client/issues",
    },
    packages=find_packages(exclude=["examples"]),
    package_data={
        "pytdbot": ["td_api.*"],
    },
    keywords=[
        "telegram",
        "tdlib",
        "bot",
        "telegram-client",
        "telegram-bot",
        "bot-api",
        "telegram-bot",
        "tdlib-python",
        "tdlib-bot",
    ],
)



__all__ = ("StopHandlers", "AuthorizationError")


class StopHandlers(Exception):
    r"""An exception to stop handlers from execution"""

    pass


class AuthorizationError(Exception):
    r"""An exception for authorization errors"""

    pass


class WebAppDataInvalid(Exception):
    r"""An exception for invalid webapp data"""

    pass


class WebAppDataOutdated(Exception):
    r"""An exception for outdated webapp data"""

    pass


class WebAppDataMismatch(Exception):
    r"""An exception for mismatched webapp data"""

    pass



from typing import List, Union

import pytdbot

from ..types import (
    ChatTypeSupergroup,
    Error,
    FormattedText,
    InputFile,
    InputFileRemote,
    InputMessageAnimation,
    InputMessageAudio,
    InputMessageContent,
    InputMessageDocument,
    InputMessageForwarded,
    InputMessagePhoto,
    InputMessageReplyTo,
    InputMessageReplyToMessage,
    InputMessageSticker,
    InputMessageText,
    InputMessageVideo,
    InputMessageVideoNote,
    InputMessageVoiceNote,
    InputTextQuote,
    InputThumbnail,
    LinkPreviewOptions,
    Message,
    MessageCopyOptions,
    MessageSelfDestructType,
    MessageSendOptions,
    ReplyMarkup,
    ReplyMarkupForceReply,
    ReplyMarkupInlineKeyboard,
    ReplyMarkupRemoveKeyboard,
    ReplyMarkupShowKeyboard,
    TextEntity,
    TextParseModeHTML,
    TextParseModeMarkdown,
)
from .td_functions import TDLibFunctions


class Methods(TDLibFunctions):
    r"""TDLib API functions class"""

    async def parseText(
        self,
        text: str,
        parse_mode: str = "html",
    ) -> Union[Error, FormattedText]:
        r"""Parses Bold, Italic, Underline, Strikethrough, Spoiler, Code, Pre, PreCode, TextUrl and MentionName entities contained in the text

        Parameters:
            text (``str``):
                The text to parse

            parse_mode (``str``):
                Text parse mode. Currently supported: markdown, markdownv2 and html. Default is "html"

        Returns:
            :class:`~pytdbot.types.FormattedText`
        """

        if not text or not parse_mode:
            return

        parse_mode = parse_mode.lower()

        if parse_mode == "markdown":
            mode = TextParseModeMarkdown(1)
        elif parse_mode == "markdownv2":
            mode = TextParseModeMarkdown(2)
        elif parse_mode == "html":
            mode = TextParseModeHTML()
        else:
            raise ValueError(
                "Invalid parse_mode. Currently supported: markdown, markdownv2, html"
            )

        return await self.parseTextEntities(text, mode)

    async def getSupergroupId(
        self,
        chat_id: int,
    ) -> Union[Error, int, None]:
        r"""Get supergroup id from chat id

        Parameters:
            chat_id (``int``):
                Chat identifier

        Returns:
            ``int``
        """

        chat_info = await self.getChat(chat_id)
        if not chat_info:
            return chat_info

        if isinstance(chat_info.type, ChatTypeSupergroup):
            return chat_info.type.supergroup_id

        return None

    async def sendTextMessage(
        self,
        chat_id: int,
        text: str,
        entities: List[TextEntity] = None,
        parse_mode: str = None,
        disable_web_page_preview: bool = False,
        url: str = None,
        force_small_media: bool = None,
        force_large_media: bool = None,
        show_above_text: bool = None,
        clear_draft: bool = False,
        disable_notification: bool = False,
        protect_content: bool = False,
        allow_paid_broadcast: bool = False,
        message_thread_id: int = 0,
        quote: InputTextQuote = None,
        reply_to: InputMessageReplyTo = None,
        reply_to_message_id: int = 0,
        reply_markup: Union[
            ReplyMarkupInlineKeyboard,
            ReplyMarkupShowKeyboard,
            ReplyMarkupForceReply,
            ReplyMarkupRemoveKeyboard,
        ] = None,
    ) -> Union[Error, Message]:
        r"""Send text message to chat

        Parameters:
            chat_id (``int``):
                Target chat

            text (``str``):
                Text to send

            entities (``list``, *optional*):
                List of ``TextEntity`` objects to parse in the text. If you want to send a text with formatting, use ```parse_mode``` instead

            parse_mode (``str``, *optional*):
                Mode for parsing entities. Default is ``markdown``

            disable_web_page_preview (``bool``, *optional*):
                Disables link previews for links in this message. Default is ``False``

            url (``str``, *optional*):
                URL to use for link preview. If empty, then the first URL found in the message text will be used. Default is ``None``

            force_small_media (``bool``, *optional*):
                True, if shown media preview must be small; ignored in secret chats or if the URL isn't explicitly specified. Default is ``None``

            force_large_media (``bool``, *optional*):
                True, if shown media preview must be large; ignored in secret chats or if the URL isn't explicitly specified. Default is ``None``

            show_above_text (``bool``, *optional*):
                True, if link preview must be shown above message text; otherwise, the link preview will be shown below the message text; ignored in secret chats. Default is ``None``

            disable_notification (``bool``, *optional*):
                If True, disable notification for the message. Default is ``None``

            clear_draft (``bool``, *optional*):
                True, if a chat message draft must be deleted. Default is ``False``

            protect_content (``bool``, *optional*):
                If True, the content of the message must be protected from forwarding and saving

            allow_paid_broadcast (``bool``, *optional*):
                Pass true to allow the message to ignore regular broadcast limits for a small fee; for bots only. Default is ``False``

            message_thread_id (``int``, *optional*):
                If not 0, a message thread identifier in which the message will be sent

            quote (:class:`~pytdbot.types.InputTextQuote`, *optional*):
                Chosen quote from the replied message; may be null if none

            reply_to (:class:`~pytdbot.types.InputMessageReplyTo`, *optional*):
                Information about the message or the story this message is replying to; may be null if none

            reply_to_message_id (``int``, *optional*):
                Identifier of the message to reply. Ignored if ``reply_to`` is specified

            reply_markup (:class:`~pytdbot.types.ReplyMarkupInlineKeyboard` | :class:`~pytdbot.types.ReplyMarkupShowKeyboard` | :class:`~pytdbot.types.ReplyMarkupForceReply` | :class:`~pytdbot.types.ReplyMarkupRemoveKeyboard`, *optional*):
                The message reply markup

        Returns:
            :class:`~pytdbot.types.Message`
        """

        parse_mode = parse_mode or self.default_parse_mode
        if entities and isinstance(entities, list):
            text = FormattedText(text=text, entities=entities)
        elif parse_mode and isinstance(parse_mode, str):
            parse = await self.parseText(text, parse_mode=parse_mode)
            if isinstance(parse, Error):
                return parse
            text = parse
        else:
            text = FormattedText(text)

        return await self.sendMessageWithContent(
            chat_id=chat_id,
            content=InputMessageText(
                text,
                link_preview_options=LinkPreviewOptions(
                    is_disabled=disable_web_page_preview,
                    url=url,
                    force_small_media=force_small_media,
                    force_large_media=force_large_media,
                    show_above_text=show_above_text,
                ),
                clear_draft=clear_draft,
            ),
            disable_notification=disable_notification,
            protect_content=protect_content,
            allow_paid_broadcast=allow_paid_broadcast,
            message_thread_id=message_thread_id,
            quote=quote,
            reply_to=reply_to,
            reply_to_message_id=reply_to_message_id,
            reply_markup=reply_markup,
        )

    async def sendAnimation(
        self,
        chat_id: int,
        animation: Union[InputFile, str],
        thumbnail: InputThumbnail = None,
        caption: str = None,
        caption_entities: list = None,
        parse_mode: str = None,
        added_sticker_file_ids: list = None,
        duration: int = 0,
        width: int = 0,
        height: int = 0,
        disable_notification: bool = False,
        protect_content: bool = False,
        allow_paid_broadcast: bool = False,
        has_spoiler: bool = False,
        message_thread_id: int = 0,
        quote: InputTextQuote = None,
        reply_to: InputMessageReplyTo = None,
        reply_to_message_id: int = 0,
        reply_markup: Union[
            ReplyMarkupInlineKeyboard,
            ReplyMarkupShowKeyboard,
            ReplyMarkupForceReply,
            ReplyMarkupRemoveKeyboard,
        ] = None,
    ) -> Union[Error, Message]:
        r"""Send animation to chat

        Parameters:
            chat_id (``int``):
                Target chat

            animation (:class:`~pytdbot.types.InputFileRemote` | :class:`~pytdbot.types.InputFileLocal` | ``str``):
                Animation to send

            thumbnail (:class:`~pytdbot.types.InputThumbnail`, *optional*):
                Thumbnail of the animation to send

            caption (``str``, *optional*):
                Animation caption

            caption_entities (``list``, *optional*):
                List of ``MessageEntity`` objects to parse in the caption. If you want to send a caption with formatting, use ``parse_mode`` instead

            parse_mode (``str``, *optional*):
                Mode for parsing entities. Default is ``markdown``

            added_sticker_file_ids (``list``, *optional*):
                File identifiers of the stickers added to the animation, if applicable

            duration (``int``, *optional*):
                Duration of the animation, in seconds

            width (``int``, *optional*):
                Width of the animation

            height (``int``, *optional*):
                Height of the animation

            disable_notification (``bool``, *optional*):
                If True, disable notification for the message. Default is ``False``

            protect_content (``bool``, *optional*):
                If True, the content of the message must be protected from forwarding and saving

            allow_paid_broadcast (``bool``, *optional*):
                Pass true to allow the message to ignore regular broadcast limits for a small fee; for bots only. Default is ``False``

            has_spoiler (``bool``, *optional*):
                True, if the photo preview must be covered by a spoiler animation; not supported in secret chats

            message_thread_id (``int``, *optional*):
                If not 0, a message thread identifier in which the message will be sent

            quote (:class:`~pytdbot.types.InputTextQuote`, *optional*):
                Chosen quote from the replied message; may be null if none

            reply_to (:class:`~pytdbot.types.InputMessageReplyTo`, *optional*):
                Information about the message or the story this message is replying to; may be null if none

            reply_to_message_id (``int``, *optional*):
                Identifier of the message to reply. Ignored if ``reply_to`` is specified

            reply_markup (:class:`~pytdbot.types.ReplyMarkupInlineKeyboard` | :class:`~pytdbot.types.ReplyMarkupShowKeyboard` | :class:`~pytdbot.types.ReplyMarkupForceReply` | :class:`~pytdbot.types.ReplyMarkupRemoveKeyboard`, *optional*):
                The message reply markup

        Returns:
            :class:`~pytdbot.types.Message`

        """

        parse_mode = parse_mode or self.default_parse_mode
        if isinstance(caption_entities, list):
            caption = FormattedText(text=caption, entities=caption_entities)
        elif parse_mode and isinstance(parse_mode, str):
            parse = await self.parseText(caption, parse_mode=parse_mode)
            if isinstance(parse, Error):
                return parse
            caption = parse
        else:
            caption = FormattedText(caption)

        if isinstance(animation, str):
            animation = InputFileRemote(animation)

        return await self.sendMessageWithContent(
            chat_id=chat_id,
            content=InputMessageAnimation(
                animation=animation,
                thumbnail=thumbnail,
                added_sticker_file_ids=added_sticker_file_ids,
                duration=duration,
                width=width,
                height=height,
                caption=caption,
                has_spoiler=has_spoiler,
            ),
            disable_notification=disable_notification,
            protect_content=protect_content,
            allow_paid_broadcast=allow_paid_broadcast,
            message_thread_id=message_thread_id,
            quote=quote,
            reply_to=reply_to,
            reply_to_message_id=reply_to_message_id,
            reply_markup=reply_markup,
        )

    async def sendAudio(
        self,
        chat_id: int,
        audio: Union[InputFile, str],
        album_cover_thumbnail: InputThumbnail = None,
        caption: str = None,
        caption_entities: list = None,
        parse_mode: str = None,
        title: str = None,
        performer: str = None,
        duration: int = 0,
        disable_notification: bool = False,
        protect_content: bool = False,
        allow_paid_broadcast: bool = False,
        message_thread_id: int = 0,
        quote: InputTextQuote = None,
        reply_to: InputMessageReplyTo = None,
        reply_to_message_id: int = 0,
        reply_markup: Union[
            ReplyMarkupInlineKeyboard,
            ReplyMarkupShowKeyboard,
            ReplyMarkupForceReply,
            ReplyMarkupRemoveKeyboard,
        ] = None,
    ) -> Union[Error, Message]:
        r"""Send audio to chat

        Parameters:
            chat_id (``int``):
                Target chat

            audio (:class:`~pytdbot.types.InputFileRemote` | :class:`~pytdbot.types.InputFileLocal` | ``str``):
                Audio to send

            album_cover_thumbnail (:class:`~pytdbot.types.InputThumbnail`, *optional*):
                Thumbnail of the album cover to be set

            caption (``str``, *optional*):
                Audio caption

            caption_entities (``list``, *optional*):
                List of ``MessageEntity`` objects to parse in the caption. If you want to send a caption with formatting, use ``parse_mode`` instead

            parse_mode (``str``, *optional*):
                Mode for parsing entities. Default is ``markdown``

            title (``str``, *optional*):
                Title of the audio

            performer (``str``, *optional*):
                Performer of the audio

            duration (``int``, *optional*):
                Duration of the audio, in seconds

            disable_notification (``bool``, *optional*):
                If True, disable notification for the message. Default is ``False``

            protect_content (``bool``, *optional*):
                If True, the content of the message must be protected from forwarding and saving

            allow_paid_broadcast (``bool``, *optional*):
                Pass true to allow the message to ignore regular broadcast limits for a small fee; for bots only. Default is ``False``

            message_thread_id (``int``, *optional*):
                If not 0, a message thread identifier in which the message will be sent

            quote (:class:`~pytdbot.types.InputTextQuote`, *optional*):
                Chosen quote from the replied message; may be null if none

            reply_to (:class:`~pytdbot.types.InputMessageReplyTo`, *optional*):
                Information about the message or the story this message is replying to; may be null if none

            reply_to_message_id (``int``, *optional*):
                Identifier of the message to reply. Ignored if ``reply_to`` is specified

            reply_markup (:class:`~pytdbot.types.ReplyMarkupInlineKeyboard` | :class:`~pytdbot.types.ReplyMarkupShowKeyboard` | :class:`~pytdbot.types.ReplyMarkupForceReply` | :class:`~pytdbot.types.ReplyMarkupRemoveKeyboard`, *optional*):
                The message reply markup


        Returns:
            :class:`~pytdbot.types.Message`

        """

        parse_mode = parse_mode or self.default_parse_mode
        if isinstance(caption_entities, list):
            caption = FormattedText(text=caption, entities=caption_entities)
        elif parse_mode and isinstance(parse_mode, str):
            parse = await self.parseText(caption, parse_mode=parse_mode)
            if isinstance(parse, Error):
                return parse
            caption = parse
        else:
            caption = FormattedText(caption)

        if isinstance(audio, str):
            audio = InputFileRemote(audio)

        return await self.sendMessageWithContent(
            chat_id=chat_id,
            content=InputMessageAudio(
                audio=audio,
                album_cover_thumbnail=album_cover_thumbnail,
                title=title,
                performer=performer,
                duration=duration,
                caption=caption,
            ),
            disable_notification=disable_notification,
            protect_content=protect_content,
            allow_paid_broadcast=allow_paid_broadcast,
            message_thread_id=message_thread_id,
            quote=quote,
            reply_to=reply_to,
            reply_to_message_id=reply_to_message_id,
            reply_markup=reply_markup,
        )

    async def sendDocument(
        self,
        chat_id: int,
        document: Union[InputFile, str],
        thumbnail: InputThumbnail = None,
        caption: str = None,
        caption_entities: list = None,
        parse_mode: str = None,
        disable_content_type_detection: bool = True,
        disable_notification: bool = False,
        protect_content: bool = False,
        allow_paid_broadcast: bool = False,
        message_thread_id: int = 0,
        quote: InputTextQuote = None,
        reply_to: InputMessageReplyTo = None,
        reply_to_message_id: int = 0,
        reply_markup: Union[
            ReplyMarkupInlineKeyboard,
            ReplyMarkupShowKeyboard,
            ReplyMarkupForceReply,
            ReplyMarkupRemoveKeyboard,
        ] = None,
    ) -> Union[Error, Message]:
        r"""Send document to chat

        Parameters:
            chat_id (``int``):
                Target chat

            document (:class:`~pytdbot.types.InputFileRemote` | :class:`~pytdbot.types.InputFileLocal` | ``str``):
                Document to send

            thumbnail (:class:`~pytdbot.types.InputThumbnail`, *optional*):
                Thumbnail of the document to be set

            caption (``str``, *optional*):
                Document caption

            caption_entities (``list``, *optional*):
                List of ``MessageEntity`` objects to parse in the caption. If you want to send a caption with formatting, use ``parse_mode`` instead

            parse_mode (``str``, *optional*):
                Mode for parsing entities. Default is ``markdown``

            disable_content_type_detection (``bool``, *optional*):
                If true, automatic file type detection will be disabled and the document will be always sent as file. Always true for files sent to secret chats

            disable_notification (``bool``, *optional*):
                If True, disable notification for the message. Default is ``False``

            protect_content (``bool``, *optional*):
                If True, the content of the message must be protected from forwarding and saving

            allow_paid_broadcast (``bool``, *optional*):
                Pass true to allow the message to ignore regular broadcast limits for a small fee; for bots only. Default is ``False``

            message_thread_id (``int``, *optional*):
                If not 0, a message thread identifier in which the message will be sent

            quote (:class:`~pytdbot.types.InputTextQuote`, *optional*):
                Chosen quote from the replied message; may be null if none

            reply_to (:class:`~pytdbot.types.InputMessageReplyTo`, *optional*):
                Information about the message or the story this message is replying to; may be null if none

            reply_to_message_id (``int``, *optional*):
                Identifier of the message to reply. Ignored if ``reply_to`` is specified

            reply_markup (:class:`~pytdbot.types.ReplyMarkupInlineKeyboard` | :class:`~pytdbot.types.ReplyMarkupShowKeyboard` | :class:`~pytdbot.types.ReplyMarkupForceReply` | :class:`~pytdbot.types.ReplyMarkupRemoveKeyboard`, *optional*):
                The message reply markup


        Returns:
            :class:`~pytdbot.types.Message`
        """

        parse_mode = parse_mode or self.default_parse_mode
        if isinstance(caption_entities, list):
            caption = FormattedText(text=caption, entities=caption_entities)
        elif parse_mode and isinstance(parse_mode, str):
            parse = await self.parseText(caption, parse_mode=parse_mode)
            if isinstance(parse, Error):
                return parse
            caption = parse
        else:
            caption = FormattedText(caption)

        if isinstance(document, str):
            document = InputFileRemote(document)

        return await self.sendMessageWithContent(
            chat_id=chat_id,
            content=InputMessageDocument(
                document=document,
                thumbnail=thumbnail,
                disable_content_type_detection=disable_content_type_detection,
                caption=caption,
            ),
            disable_notification=disable_notification,
            protect_content=protect_content,
            allow_paid_broadcast=allow_paid_broadcast,
            message_thread_id=message_thread_id,
            quote=quote,
            reply_to=reply_to,
            reply_to_message_id=reply_to_message_id,
            reply_markup=reply_markup,
        )

    async def sendPhoto(
        self,
        chat_id: int,
        photo: Union[InputFile, str],
        thumbnail: InputThumbnail = None,
        caption: str = None,
        caption_entities: list = None,
        parse_mode: str = None,
        added_sticker_file_ids: list = None,
        width: int = 0,
        height: int = 0,
        self_destruct_type: MessageSelfDestructType = None,
        disable_notification: bool = False,
        protect_content: bool = False,
        allow_paid_broadcast: bool = False,
        has_spoiler: bool = False,
        message_thread_id: int = 0,
        quote: InputTextQuote = None,
        reply_to: InputMessageReplyTo = None,
        reply_to_message_id: int = 0,
        reply_markup: Union[
            ReplyMarkupInlineKeyboard,
            ReplyMarkupShowKeyboard,
            ReplyMarkupForceReply,
            ReplyMarkupRemoveKeyboard,
        ] = None,
    ) -> Union[Error, Message]:
        r"""Send photo to chat

        Parameters:
            chat_id (``int``):
                Target chat

            photo (:class:`~pytdbot.types.InputFileRemote` | :class:`~pytdbot.types.InputFileLocal` | ``str``):
                Photo to send

            thumbnail (:class:`~pytdbot.types.InputThumbnail`, *optional*):
                Thumbnail of the photo to be set

            caption (``str``, *optional*):
                Photo caption

            caption_entities (``list``, *optional*):
                List of ``MessageEntity`` objects to parse in the caption. If you want to send a caption with formatting, use ``parse_mode`` instead

            parse_mode (``str``, *optional*):
                Mode for parsing entities. Default is ``markdown``

            added_sticker_file_ids (``list``, *optional*):
                List of file identifiers of added stickers

            width (``int``, *optional*):
                Photo width

            height (``int``, *optional*):
                Photo height

            self_destruct_type (:class:`~pytdbot.types.MessageSelfDestructType`, *optional*):
                Photo self-destruct type; pass null if none; private chats only

            disable_notification (``bool``, *optional*):
                If True, disable notification for the message. Default is ``False``

            protect_content (``bool``, *optional*):
                If True, the content of the message must be protected from forwarding and saving

            allow_paid_broadcast (``bool``, *optional*):
                Pass true to allow the message to ignore regular broadcast limits for a small fee; for bots only. Default is ``False``

            has_spoiler (``bool``, *optional*):
                True, if the photo preview must be covered by a spoiler animation; not supported in secret chats

            message_thread_id (``int``, *optional*):
                If not 0, a message thread identifier in which the message will be sent

            quote (:class:`~pytdbot.types.InputTextQuote`, *optional*):
                Chosen quote from the replied message; may be null if none

            reply_to (:class:`~pytdbot.types.InputMessageReplyTo`, *optional*):
                Information about the message or the story this message is replying to; may be null if none

            reply_to_message_id (``int``, *optional*):
                Identifier of the message to reply. Ignored if ``reply_to`` is specified

            reply_markup (:class:`~pytdbot.types.ReplyMarkupInlineKeyboard` | :class:`~pytdbot.types.ReplyMarkupShowKeyboard` | :class:`~pytdbot.types.ReplyMarkupForceReply` | :class:`~pytdbot.types.ReplyMarkupRemoveKeyboard`, *optional*):
                The message reply markup


        Returns:
            :class:`~pytdbot.types.Message`
        """

        parse_mode = parse_mode or self.default_parse_mode
        if isinstance(caption_entities, list):
            caption = FormattedText(text=caption, entities=caption_entities)
        elif parse_mode and isinstance(parse_mode, str):
            parse = await self.parseText(caption, parse_mode=parse_mode)
            if isinstance(parse, Error):
                return parse
            caption = parse
        else:
            caption = FormattedText(caption)

        if isinstance(photo, str):
            photo = InputFileRemote(photo)

        return await self.sendMessageWithContent(
            chat_id=chat_id,
            content=InputMessagePhoto(
                photo=photo,
                thumbnail=thumbnail,
                added_sticker_file_ids=added_sticker_file_ids,
                self_destruct_type=self_destruct_type,
                width=width,
                height=height,
                caption=caption,
                has_spoiler=has_spoiler,
            ),
            disable_notification=disable_notification,
            protect_content=protect_content,
            allow_paid_broadcast=allow_paid_broadcast,
            message_thread_id=message_thread_id,
            quote=quote,
            reply_to=reply_to,
            reply_to_message_id=reply_to_message_id,
            reply_markup=reply_markup,
        )

    async def sendVideo(
        self,
        chat_id: int,
        video: Union[InputFile, str],
        thumbnail: InputThumbnail = None,
        caption: str = None,
        caption_entities: list = None,
        parse_mode: str = None,
        added_sticker_file_ids: list = None,
        supports_streaming: bool = None,
        duration: int = 0,
        width: int = 0,
        height: int = 0,
        self_destruct_type: MessageSelfDestructType = None,
        disable_notification: bool = False,
        protect_content: bool = False,
        allow_paid_broadcast: bool = False,
        has_spoiler: bool = False,
        message_thread_id: int = 0,
        quote: InputTextQuote = None,
        reply_to: InputMessageReplyTo = None,
        reply_to_message_id: int = 0,
        reply_markup: Union[
            ReplyMarkupInlineKeyboard,
            ReplyMarkupShowKeyboard,
            ReplyMarkupForceReply,
            ReplyMarkupRemoveKeyboard,
        ] = None,
    ) -> Union[Error, Message]:
        r"""Send video to chat

        Parameters:
            chat_id (``int``):
                Target chat

            video (:class:`~pytdbot.types.InputFileRemote` | :class:`~pytdbot.types.InputFileLocal` | ``str``):
                Video to send

            thumbnail (:class:`~pytdbot.types.InputThumbnail`, *optional*):
                Thumbnail of the video to be set

            caption (``str``, *optional*):
                Video caption

            caption_entities (``list``, *optional*):
                List of ``MessageEntity`` objects to parse in the caption. If you want to send a caption with formatting, use ``parse_mode`` instead

            parse_mode (``str``, *optional*):
                Mode for parsing entities. Default is ``markdown``

            added_sticker_file_ids (``list``, *optional*):
                List of file identifiers of added stickers

            supports_streaming (``bool``, *optional*):
                True, if the video should be tried to be streamed

            duration (``int``, *optional*):
                Duration of sent video in seconds

            width (``int``, *optional*):
                Video width

            height (``int``, *optional*):
                Video height

            self_destruct_type (:class:`~pytdbot.types.MessageSelfDestructType`, *optional*):
                Video self-destruct type; pass null if none; private chats only

            disable_notification (``bool``, *optional*):
                If True, disable notification for the message. Default is ``False``

            protect_content (``bool``, *optional*):
                If True, the content of the message must be protected from forwarding and saving

            allow_paid_broadcast (``bool``, *optional*):
                Pass true to allow the message to ignore regular broadcast limits for a small fee; for bots only. Default is ``False``

            has_spoiler (``bool``, *optional*):
                True, if the photo preview must be covered by a spoiler animation; not supported in secret chats

            message_thread_id (``int``, *optional*):
                If not 0, a message thread identifier in which the message will be sent

            quote (:class:`~pytdbot.types.InputTextQuote`, *optional*):
                Chosen quote from the replied message; may be null if none

            reply_to (:class:`~pytdbot.types.InputMessageReplyTo`, *optional*):
                Information about the message or the story this message is replying to; may be null if none

            reply_to_message_id (``int``, *optional*):
                Identifier of the message to reply. Ignored if ``reply_to`` is specified

            reply_markup (:class:`~pytdbot.types.ReplyMarkupInlineKeyboard` | :class:`~pytdbot.types.ReplyMarkupShowKeyboard` | :class:`~pytdbot.types.ReplyMarkupForceReply` | :class:`~pytdbot.types.ReplyMarkupRemoveKeyboard`, *optional*):
                The message reply markup


        Returns:
            :class:`~pytdbot.types.Message`
        """

        parse_mode = parse_mode or self.default_parse_mode
        if isinstance(caption_entities, list):
            caption = FormattedText(text=caption, entities=caption_entities)
        elif parse_mode and isinstance(parse_mode, str):
            parse = await self.parseText(caption, parse_mode=parse_mode)
            if isinstance(parse, Error):
                return parse
            caption = parse
        else:
            caption = FormattedText(caption)

        if isinstance(video, str):
            video = InputFileRemote(video)

        return await self.sendMessageWithContent(
            chat_id=chat_id,
            content=InputMessageVideo(
                video=video,
                thumbnail=thumbnail,
                added_sticker_file_ids=added_sticker_file_ids,
                duration=duration,
                width=width,
                height=height,
                supports_streaming=supports_streaming,
                caption=caption,
                self_destruct_type=self_destruct_type,
                has_spoiler=has_spoiler,
            ),
            disable_notification=disable_notification,
            protect_content=protect_content,
            allow_paid_broadcast=allow_paid_broadcast,
            message_thread_id=message_thread_id,
            quote=quote,
            reply_to=reply_to,
            reply_to_message_id=reply_to_message_id,
            reply_markup=reply_markup,
        )

    async def sendVideoNote(
        self,
        chat_id: int,
        video_note: Union[InputFile, str],
        thumbnail: InputThumbnail = None,
        duration: int = 0,
        length: int = 0,
        disable_notification: bool = False,
        protect_content: bool = False,
        allow_paid_broadcast: bool = False,
        message_thread_id: int = 0,
        quote: InputTextQuote = None,
        reply_to: InputMessageReplyTo = None,
        reply_to_message_id: int = 0,
        reply_markup: Union[
            ReplyMarkupInlineKeyboard,
            ReplyMarkupShowKeyboard,
            ReplyMarkupForceReply,
            ReplyMarkupRemoveKeyboard,
        ] = None,
    ) -> Union[Error, Message]:
        r"""Send video note to chat

        Parameters:
            chat_id (``int``):
                Target chat

            video_note (:class:`~pytdbot.types.InputFileRemote` | :class:`~pytdbot.types.InputFileLocal` | ``str``):
                Video note to send

            thumbnail (:class:`~pytdbot.types.InputThumbnail`, *optional*):
                Thumbnail of the video note to be set

            duration (``int``, *optional*):
                Duration of sent video note in seconds

            length (``int``, *optional*):
                Video width and height

            disable_notification (``bool``, *optional*):
                If True, disable notification for the message. Default is ``False``

            protect_content (``bool``, *optional*):
                If True, the content of the message must be protected from forwarding and saving

            allow_paid_broadcast (``bool``, *optional*):
                Pass true to allow the message to ignore regular broadcast limits for a small fee; for bots only. Default is ``False``

            message_thread_id (``int``, *optional*):
                If not 0, a message thread identifier in which the message will be sent

            quote (:class:`~pytdbot.types.InputTextQuote`, *optional*):
                Chosen quote from the replied message; may be null if none

            reply_to (:class:`~pytdbot.types.InputMessageReplyTo`, *optional*):
                Information about the message or the story this message is replying to; may be null if none

            reply_to_message_id (``int``, *optional*):
                Identifier of the message to reply. Ignored if ``reply_to`` is specified

            reply_markup (:class:`~pytdbot.types.ReplyMarkupInlineKeyboard` | :class:`~pytdbot.types.ReplyMarkupShowKeyboard` | :class:`~pytdbot.types.ReplyMarkupForceReply` | :class:`~pytdbot.types.ReplyMarkupRemoveKeyboard`, *optional*):
                The message reply markup


        Returns:
            :class:`~pytdbot.types.Message`

        """

        if isinstance(video_note, str):
            video_note = InputFileRemote(video_note)

        return await self.sendMessageWithContent(
            chat_id=chat_id,
            content=InputMessageVideoNote(
                video_note=video_note,
                thumbnail=thumbnail,
                duration=duration,
                length=length,
            ),
            disable_notification=disable_notification,
            protect_content=protect_content,
            allow_paid_broadcast=allow_paid_broadcast,
            message_thread_id=message_thread_id,
            quote=quote,
            reply_to=reply_to,
            reply_to_message_id=reply_to_message_id,
            reply_markup=reply_markup,
        )

    async def sendVoice(
        self,
        chat_id: int,
        voice: Union[InputFile, str],
        caption: str = None,
        caption_entities: list = None,
        parse_mode: str = None,
        duration: int = 0,
        waveform: bytes = None,
        disable_notification: bool = False,
        protect_content: bool = False,
        allow_paid_broadcast: bool = False,
        message_thread_id: int = 0,
        quote: InputTextQuote = None,
        reply_to: InputMessageReplyTo = None,
        reply_to_message_id: int = 0,
        reply_markup: Union[
            ReplyMarkupInlineKeyboard,
            ReplyMarkupShowKeyboard,
            ReplyMarkupForceReply,
            ReplyMarkupRemoveKeyboard,
        ] = None,
    ) -> Union[Error, Message]:
        r"""Send voice to chat

        Parameters:
            chat_id (``int``):
                Target chat

            voice (:class:`~pytdbot.types.InputFileRemote` | :class:`~pytdbot.types.InputFileLocal` | ``str``):
                Voice to send

            caption (``str``, *optional*):
                Voice caption

            caption_entities (``list``, *optional*):
                List of ``MessageEntity`` objects to parse in the caption. If you want to send a caption without parsing entities, use ``parse_mode`` instead

            parse_mode (``str``, *optional*):
                Parse mode for the caption. Default is None

            duration (``int``, *optional*):
                Duration of sent voice in seconds

            waveform (`bytes`, *optional*):
                Waveform representation of the voice note, in 5-bit format

            disable_notification (``bool``, *optional*):
                If True, disable notification for the message. Default is ``False``

            protect_content (``bool``, *optional*):
                If True, the content of the message must be protected from forwarding and saving

            allow_paid_broadcast (``bool``, *optional*):
                Pass true to allow the message to ignore regular broadcast limits for a small fee; for bots only. Default is ``False``

            message_thread_id (``int``, *optional*):
                If not 0, a message thread identifier in which the message will be sent

            quote (:class:`~pytdbot.types.InputTextQuote`, *optional*):
                Chosen quote from the replied message; may be null if none

            reply_to (:class:`~pytdbot.types.InputMessageReplyTo`, *optional*):
                Information about the message or the story this message is replying to; may be null if none

            reply_to_message_id (``int``, *optional*):
                Identifier of the message to reply. Ignored if ``reply_to`` is specified

            reply_markup (:class:`~pytdbot.types.ReplyMarkupInlineKeyboard` | :class:`~pytdbot.types.ReplyMarkupShowKeyboard` | :class:`~pytdbot.types.ReplyMarkupForceReply` | :class:`~pytdbot.types.ReplyMarkupRemoveKeyboard`, *optional*):
                The message reply markup


        Returns:
            :class:`~pytdbot.types.Message`
        """

        parse_mode = parse_mode or self.default_parse_mode
        if isinstance(caption_entities, list):
            caption = FormattedText(text=caption, entities=caption_entities)
        elif parse_mode and isinstance(parse_mode, str):
            parse = await self.parseText(caption, parse_mode=parse_mode)
            if isinstance(parse, Error):
                return parse
            caption = parse
        else:
            caption = FormattedText(caption)

        if isinstance(voice, str):
            voice = InputFileRemote(voice)

        return await self.sendMessageWithContent(
            chat_id=chat_id,
            content=InputMessageVoiceNote(
                voice_note=voice,
                waveform=waveform,
                duration=duration,
                caption=caption,
            ),
            disable_notification=disable_notification,
            protect_content=protect_content,
            allow_paid_broadcast=allow_paid_broadcast,
            message_thread_id=message_thread_id,
            quote=quote,
            reply_to=reply_to,
            reply_to_message_id=reply_to_message_id,
            reply_markup=reply_markup,
        )

    async def sendSticker(
        self,
        chat_id: int,
        sticker: Union[InputFile, str],
        emoji: str = None,
        thumbnail: InputThumbnail = None,
        width: int = 0,
        height: int = 0,
        disable_notification: bool = False,
        protect_content: bool = False,
        allow_paid_broadcast: bool = False,
        message_thread_id: int = 0,
        quote: InputTextQuote = None,
        reply_to: InputMessageReplyTo = None,
        reply_to_message_id: int = 0,
        reply_markup: Union[
            ReplyMarkupInlineKeyboard,
            ReplyMarkupShowKeyboard,
            ReplyMarkupForceReply,
            ReplyMarkupRemoveKeyboard,
        ] = None,
    ) -> Union[Error, Message]:
        r"""Send sticker to chat

        Parameters:
            chat_id (``int``):
                Target chat

            sticker (:class:`~pytdbot.types.InputFileRemote` | :class:`~pytdbot.types.InputFileLocal` | ``str``):
                Sticker to send

            emoji (``str``, *optional*):
                Emoji associated with the sticker

            thumbnail (:class:`~pytdbot.types.InputThumbnail`, *optional*):
                Sticker thumbnail

            width (``int``, *optional*):
                Sticker width

            height (``int``, *optional*):
                Sticker height

            disable_notification (``bool``, *optional*):
                If True, disable notification for the message. Default is ``False``

            protect_content (``bool``, *optional*):
                If True, the content of the message must be protected from forwarding and saving

            allow_paid_broadcast (``bool``, *optional*):
                Pass true to allow the message to ignore regular broadcast limits for a small fee; for bots only. Default is ``False``

            message_thread_id (``int``, *optional*):
                If not 0, a message thread identifier in which the message will be sent

            quote (:class:`~pytdbot.types.InputTextQuote`, *optional*):
                Chosen quote from the replied message; may be null if none

            reply_to (:class:`~pytdbot.types.InputMessageReplyTo`, *optional*):
                Information about the message or the story this message is replying to; may be null if none

            reply_to_message_id (``int``, *optional*):
                Identifier of the message to reply. Ignored if ``reply_to`` is specified

            reply_markup (:class:`~pytdbot.types.ReplyMarkupInlineKeyboard` | :class:`~pytdbot.types.ReplyMarkupShowKeyboard` | :class:`~pytdbot.types.ReplyMarkupForceReply` | :class:`~pytdbot.types.ReplyMarkupRemoveKeyboard`, *optional*):
                The message reply markup


        Returns:
            :class:`~pytdbot.types.Message`
        """

        if isinstance(sticker, str):
            sticker = InputFileRemote(sticker)

        return await self.sendMessageWithContent(
            chat_id=chat_id,
            content=InputMessageSticker(
                sticker=sticker,
                thumbnail=thumbnail,
                width=width,
                height=height,
                emoji=emoji,
            ),
            disable_notification=disable_notification,
            protect_content=protect_content,
            allow_paid_broadcast=allow_paid_broadcast,
            message_thread_id=message_thread_id,
            quote=quote,
            reply_to=reply_to,
            reply_to_message_id=reply_to_message_id,
            reply_markup=reply_markup,
        )

    async def sendCopy(
        self,
        chat_id: int,
        from_chat_id: int,
        message_id: int,
        in_game_share: bool = None,
        replace_caption: bool = None,
        new_caption: str = None,
        new_caption_entities: list = None,
        parse_mode: str = None,
        disable_notification: bool = False,
        protect_content: bool = False,
        allow_paid_broadcast: bool = False,
        message_thread_id: int = 0,
        quote: InputTextQuote = None,
        reply_to: InputMessageReplyTo = None,
        reply_to_message_id: int = 0,
    ) -> Union[Error, Message]:
        r"""Copy message to chat

        Parameters:
            chat_id (``int``):
                Target chat

            from_chat_id (``int``):
                Identifier for the chat this forwarded message came from

            message_id (``int``):
                Identifier of the message to forward

            in_game_share (``bool``, *optional*):
                True, if a game message is being shared from a launched game; applies only to game messages

            replace_caption (``bool``, *optional*):
                True, if media caption of the message copy needs to be replaced

            new_caption (``str``, *optional*):
                New caption of the message copy

            new_caption_entities (``list``, *optional*):
                List of ``MessageEntity`` objects representing entities in the new caption

            parse_mode (``str``, *optional*):
                Mode for parsing entities in the new caption

            disable_notification (``bool``, *optional*):
                If True, disable notification for the message. Default is ``False``

            protect_content (``bool``, *optional*):
                If True, the content of the message must be protected from forwarding and saving

            allow_paid_broadcast (``bool``, *optional*):
                Pass true to allow the message to ignore regular broadcast limits for a small fee; for bots only. Default is ``False``

            message_thread_id (``int``, *optional*):
                If not 0, a message thread identifier in which the message will be sent

            quote (:class:`~pytdbot.types.InputTextQuote`, *optional*):
                Chosen quote from the replied message; may be null if none

            reply_to (:class:`~pytdbot.types.InputMessageReplyTo`, *optional*):
                Information about the message or the story this message is replying to; may be null if none

            reply_to_message_id (``int``, *optional*):
                Identifier of the message to reply. Ignored if ``reply_to`` is specified


        Returns:
            :class:`~pytdbot.types.Message`
        """

        parse_mode = parse_mode or self.default_parse_mode
        if new_caption_entities and isinstance(new_caption_entities, list):
            caption = FormattedText(text=new_caption, entities=new_caption_entities)
        elif parse_mode and isinstance(parse_mode, str):
            parse = await self.parseText(new_caption, parse_mode=parse_mode)
            if isinstance(parse, Error):
                return parse
            caption = parse
        else:
            caption = FormattedText(new_caption)

        return await self.sendMessageWithContent(
            chat_id=chat_id,
            content=InputMessageForwarded(
                from_chat_id=from_chat_id,
                message_id=message_id,
                in_game_share=in_game_share,
                copy_options=MessageCopyOptions(
                    send_copy=True, replace_caption=replace_caption, new_caption=caption
                ),
            ),
            disable_notification=disable_notification,
            protect_content=protect_content,
            allow_paid_broadcast=allow_paid_broadcast,
            message_thread_id=message_thread_id,
            quote=quote,
            reply_to=reply_to,
            reply_to_message_id=reply_to_message_id,
        )

    async def forwardMessage(
        self,
        chat_id: int,
        from_chat_id: int,
        message_id: int,
        in_game_share: bool = False,
        disable_notification: bool = False,
    ):
        r"""Forward message to chat

        Parameters:
            chat_id (``int``):
                Target chat

            from_chat_id (``int``):
                Identifier for the chat this forwarded message came from

            message_id (``int``):
                Identifier of the message to forward

            in_game_share (``bool``, *optional*):
                True, if a game message is being shared from a launched game; applies only to game messages

            disable_notification (``bool``, *optional*):
                If True, disable notification for the message. Default is ``False``

        Returns:
            :class:`~pytdbot.types.Message`
        """

        return await self.sendMessageWithContent(
            chat_id=chat_id,
            content=InputMessageForwarded(
                from_chat_id=from_chat_id,
                message_id=message_id,
                in_game_share=in_game_share,
            ),
            disable_notification=disable_notification,
        )

    async def editTextMessage(
        self,
        chat_id: int,
        message_id: int,
        text: str,
        parse_mode: str = None,
        entities: list = None,
        disable_web_page_preview: bool = False,
        url: str = None,
        force_small_media: bool = None,
        force_large_media: bool = None,
        show_above_text: bool = None,
        reply_markup: ReplyMarkup = None,
    ) -> Union[Error, Message]:
        r"""Edit text message

        Parameters:
            chat_id (``int``):
                Chat identifier

            message_id (``int``):
                Message identifier in the chat specified in chat_id

            text (``str``):
                New text of the message

            parse_mode (``str``, *optional*):
                Mode for parsing entities in the message text

            entities (``list``, *optional*):
                List of ``MessageEntity`` objects representing entities that may appear in the message text

            disable_web_page_preview (``bool``, *optional*):
                Disables link previews for links in this message. Default is ``False``

            url (``str``, *optional*):
                URL to use for link preview. If empty, then the first URL found in the message text will be used. Default is ``None``

            force_small_media (``bool``, *optional*):
                True, if shown media preview must be small; ignored in secret chats or if the URL isn't explicitly specified. Default is ``None``

            force_large_media (``bool``, *optional*):
                True, if shown media preview must be large; ignored in secret chats or if the URL isn't explicitly specified. Default is ``None``

            show_above_text (``bool``, *optional*):
                True, if link preview must be shown above message text; otherwise, the link preview will be shown below the message text; ignored in secret chats. Default is ``None``

            reply_markup (:class:`~pytdbot.types.ReplyMarkupInlineKeyboard` | :class:`~pytdbot.types.ReplyMarkupShowKeyboard` | :class:`~pytdbot.types.ReplyMarkupForceReply` | :class:`~pytdbot.types.ReplyMarkupRemoveKeyboard`, *optional*):
                The message reply markup


        Returns:
            :class:`~pytdbot.types.Message`
        """

        if not self.use_message_database:
            load_message = await self.getMessage(chat_id, message_id)
            if isinstance(load_message, Error):
                return load_message

        parse_mode = parse_mode or self.default_parse_mode
        if entities and isinstance(entities, list):
            text = FormattedText(text=text, entities=entities)
        elif parse_mode and isinstance(parse_mode, str):
            parse = await self.parseText(text, parse_mode=parse_mode)
            if isinstance(parse, Error):
                return parse
            text = parse
        else:
            text = FormattedText(text)

        return await self.editMessageText(
            chat_id=chat_id,
            message_id=message_id,
            reply_markup=reply_markup
            if isinstance(reply_markup, ReplyMarkup)
            else None,
            input_message_content=InputMessageText(
                text=text,
                link_preview_options=LinkPreviewOptions(
                    is_disabled=disable_web_page_preview,
                    url=url,
                    force_small_media=force_small_media,
                    force_large_media=force_large_media,
                    show_above_text=show_above_text,
                ),
            ),
        )

    async def sendMessageWithContent(
        self,
        chat_id: int,
        content: InputMessageContent,
        disable_notification: bool = False,
        protect_content: bool = False,
        allow_paid_broadcast: bool = False,
        message_thread_id: int = 0,
        quote: InputTextQuote = None,
        reply_to: InputMessageReplyTo = None,
        reply_to_message_id: int = 0,
        reply_markup: Union[
            ReplyMarkupInlineKeyboard,
            ReplyMarkupShowKeyboard,
            ReplyMarkupForceReply,
            ReplyMarkupRemoveKeyboard,
        ] = None,
    ):
        if isinstance(reply_to_message_id, int) and reply_to_message_id > 0:
            reply_to = InputMessageReplyToMessage(
                message_id=reply_to_message_id, quote=quote
            )

        if self.load_messages_before_reply and isinstance(
            reply_to, InputMessageReplyToMessage
        ):
            await self.getMessage(chat_id=chat_id, message_id=reply_to.message_id)

        res = await self.sendMessage(
            chat_id=chat_id,
            message_thread_id=message_thread_id,
            reply_to=reply_to,
            options=MessageSendOptions(
                disable_notification=disable_notification,
                protect_content=protect_content,
                allow_paid_broadcast=allow_paid_broadcast,
            ),
            reply_markup=reply_markup
            if isinstance(reply_markup, ReplyMarkup)
            else None,
            input_message_content=content,
        )
        if isinstance(res, Error):
            return res

        return await self._create_request_future(None, f"{res.chat_id}:{res.id}")



import json
from base64 import b64encode

from .. import types, utils


def obj_to_json(obj, **kwargs):
    return json.dumps(obj_to_dict(obj), **kwargs)


def obj_to_dict(obj):
    if hasattr(obj, "to_dict"):
        return obj_to_dict(obj.to_dict())
    elif isinstance(obj, list):
        return [obj_to_dict(item) for item in obj]
    elif isinstance(obj, dict):
        return {key: obj_to_dict(value) for key, value in obj.items()}
    elif isinstance(obj, bytes):
        return b64encode(obj).decode("utf-8")
    else:
        return obj


def dict_to_obj(dict_obj, client=None):
    if isinstance(dict_obj, dict):
        if "@type" in dict_obj:
            obj = getattr(types, utils.to_camel_case(dict_obj["@type"])).from_dict(
                {key: dict_to_obj(value, client) for key, value in dict_obj.items()}
            )
            if client:
                obj._client = client
            return obj
        else:
            return {key: dict_to_obj(value, client) for key, value in dict_obj.items()}
    elif isinstance(dict_obj, list):
        return [dict_to_obj(item, client) for item in dict_obj]
    else:
        return dict_obj



import binascii
import os

RETRY_AFTER_PREFEX = "Too Many Requests: retry after "


def to_camel_case(input_str: str, delimiter: str = ".", is_class: bool = True) -> str:
    if not input_str:
        return ""

    parts = input_str.split(delimiter)
    camel_case_str = ""

    for i, part in enumerate(parts):
        if i > 0:
            camel_case_str += part[0].upper() + part[1:]
        else:
            camel_case_str += part

    if camel_case_str:
        camel_case_str = (
            camel_case_str[0].upper() if is_class else camel_case_str[0].lower()
        ) + camel_case_str[1:]

    return camel_case_str


def create_extra_id(bytes_size: int = 9):
    return binascii.hexlify(os.urandom(bytes_size)).decode()


def get_bot_id_from_token(token: str) -> str:
    if len(token) > 80:
        return ""
    return token.split(":")[0] if ":" in token else ""


def get_retry_after_time(error_message: str) -> int:
    r"""Get the retry after time from flood wait error message

    Parameters:
        error_message (``str``):
            The returned error message from TDLib

    Returns:
        py:class:`int`
    """

    try:
        return int(error_message.removeprefix(RETRY_AFTER_PREFEX))
    except Exception:
        return 0



try:
    import tdjson
except ImportError:
    tdjson = None

import sys
from ctypes import CDLL, c_char_p, c_double, c_int
from logging import getLogger
from typing import Union

from ..utils import JSON_ENCODER, json_dumps, json_loads

logger = getLogger(__name__)


class TdJson:
    def __init__(self, lib_path: str = None, verbosity: int = 2) -> None:
        """TdJson client

        Parameters:
            lib_path (``str``, optional):
                Path to shared library; if ``None`` then [`tdjson`](https://github.com/AYMENJD/tdjson) binding will be used. Default is ``None``

            verbosity (``int``, optional):
                TDLib verbosity level. Default is ``2``

        Raises:
            :py:class:``ValueError``: If library not found
        """

        self._build_client(lib_path, verbosity)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    def _build_client(self, lib_path: str, verbosity: int) -> None:
        """Build TdJson client

        Parameters:
            lib_path (``str``):
                Path to shared library

            verbosity (``int``):
                TDLib verbosity level
        """

        self.using_binding = False

        if lib_path is None:
            if not tdjson:
                raise ValueError(
                    f"tdjson binding not found. Try install using: `{sys.executable} -m pip install --upgrade tdjson`"
                )

            # Use tdjson binding that already include TDLib
            self._td_create_client_id = tdjson.td_create_client_id
            self._td_send = tdjson.td_send
            self._td_receive = tdjson.td_receive
            self._td_execute = tdjson.td_execute

            self.using_binding = True

            logger.info(f"Using tdjson binding {tdjson.__version__}")
        else:
            if not lib_path:
                raise ValueError(
                    "Could not find TDLib, provide full path to libtdjson.so in lib_path"
                )

            logger.info(f"Initializing TdJson client with library: {lib_path}")

            self._tdjson = CDLL(lib_path)

            # load TDLib functions from shared library
            self._td_create_client_id = self._tdjson.td_create_client_id
            self._td_create_client_id.restype = c_int
            self._td_create_client_id.argtypes = []

            self._td_receive = self._tdjson.td_receive
            self._td_receive.restype = c_char_p
            self._td_receive.argtypes = [c_double]

            self._td_send = self._tdjson.td_send
            self._td_send.restype = None
            self._td_send.argtypes = [c_int, c_char_p]

            self._td_execute = self._tdjson.td_execute
            self._td_execute.restype = c_char_p
            self._td_execute.argtypes = [c_char_p]

        td_version, td_commit_hash = (
            self.execute({"@type": "getOption", "name": "version"}),
            self.execute({"@type": "getOption", "name": "commit_hash"}),
        )

        logger.info(
            f"Using TDLib {td_version['value']} ({td_commit_hash['value'][:9]}) with {JSON_ENCODER} encoder"
        )

        if isinstance(verbosity, int):
            res = self.execute(
                {"@type": "setLogVerbosityLevel", "new_verbosity_level": verbosity}
            )

            if res["@type"] == "error":
                logger.error("Can't set log level: {}".format(res["message"]))

    def create_client_id(self) -> int:
        """Returns an opaque identifier of a new TDLib instance"""
        return self._td_create_client_id()

    def receive(self, timeout: float = 2.0) -> Union[None, dict]:
        """Receives incoming updates and results from TDLib

        Parameters:
            timeout (``float``, *optional*):
                The maximum number of seconds allowed to wait for new data. Default is ``2.0``

        Returns:
            :py:class:``dict``: An incoming update or result to a request. If no data is received, ``None`` is returned
        """

        if res := self._td_receive(
            timeout if self.using_binding else c_double(timeout)
        ):
            return json_loads(res)

    def send(self, client_id: int, data: dict) -> None:
        """Sends a request to TDLib

        Parameters:
            client_id (``int``):
                TDLib Client identifier

            data (``dict``):
                Request to be sent
        """

        if client_id is None:
            raise ValueError("client_id is required")

        self._td_send(
            client_id,
            json_dumps(
                data,
                encode=not self.using_binding,
                null_terminated=not self.using_binding,
            ),
        )

    def execute(self, data: dict) -> Union[None, dict]:
        """Executes a TDLib request

        Parameters:
            data (``dict``): The request to be executed

        Returns:
            :py:class:``dict``: The result of the request
        """

        if res := self._td_execute(
            json_dumps(
                data,
                encode=not self.using_binding,
                null_terminated=not self.using_binding,
            )
        ):
            return json_loads(res)



from . import types, utils, filters, exception
from .tdjson import TdJson
from .client_manager import ClientManager
from .client import Client

__all__ = [
    "types",
    "utils",
    "filters",
    "exception",
    "TdJson",
    "ClientManager",
    "Client",
]

__version__ = "0.9.7"
__copyright__ = "Copyright (c) 2022-2025 Pytdbot, AYMENJD"
__license__ = "MIT License"

VERSION = __version__



import pytdbot
from typing import Literal, Union


class ServerStats:
    r"""Describes TDLib Server stats

    Parameters:
        my_id (:class:`int`):
            Identifier of the current user

        uptime (:class:`int`):
            Server uptime in seconds

        updates_count (:class:`int`):
            Total number of received updates

        requests_count (:class:`int`):
            Total number of sent requests
    """

    def __init__(
        self,
        my_id: int = 0,
        uptime: int = 0,
        updates_count: int = 0,
        requests_count: int = 0,
    ) -> None:
        self.my_id = my_id
        r"""Identifier of the current user"""
        self.uptime = uptime
        r"""Server uptime in seconds"""
        self.updates_count = updates_count
        r"""Total number of received updates"""
        self.requests_count = requests_count
        r"""Total number of sent requests"""

    def __str__(self):
        return str(pytdbot.utils.obj_to_json(self, indent=4))

    def getType(self) -> Literal["serverStats"]:
        return "serverStats"

    def getClass(self) -> Literal["ServerStats"]:
        return "ServerStats"

    def to_dict(self) -> dict:
        return {
            "@type": self.getType(),
            "my_id": self.my_id,
            "uptime": self.uptime,
            "updates_count": self.updates_count,
            "requests_count": self.requests_count,
        }

    @classmethod
    def from_dict(cls, data: dict) -> Union["ServerStats", None]:
        if data:
            data_class = cls()
            data_class.my_id = data.get("my_id", None)
            data_class.uptime = data.get("uptime", None)
            data_class.updates_count = data.get("updates_count", None)
            data_class.requests_count = data.get("requests_count", None)

        return data_class
